import os
from string import Template

from instana import eum_api_key as global_eum_api_key
from .singletons import tracer
from instana.log import logger

# Usage:
#
# from instana.helpers import eum_snippet
# meta_kvs = { 'userId': user.id }
# eum_snippet(meta=meta_kvs)


def eum_snippet(trace_id=None, eum_api_key=None, meta=None):
    """
    This method has been deprecated and will be removed in a future version.

    @param trace_id [optional] the trace ID to insert into the EUM string
    @param eum_api_key [optional] the EUM API key from your Instana dashboard
    @param meta [optional] optional additional KVs you want reported with the
                EUM metrics

    @return string
    """
    return ''


def eum_test_snippet(trace_id=None, eum_api_key=None, meta=None):
    """
    This method has been deprecated and will be removed in a future version.

    @param trace_id [optional] the trace ID to insert into the EUM string
    @param eum_api_key [optional] the EUM API key from your Instana dashboard
    @param meta [optional] optional additional KVs you want reported with the
                EUM metrics

    @return string
    """
    return ''



import os
import sys
import opentracing

from .log import logger
from .tracer import InstanaTracer
from .autoprofile.profiler import Profiler

agent = None
tracer = None
profiler = None
span_recorder = None

# Detect the environment where we are running ahead of time
aws_env = os.environ.get("AWS_EXECUTION_ENV", "")
env_is_test = "INSTANA_TEST" in os.environ
env_is_aws_fargate = aws_env == "AWS_ECS_FARGATE"
env_is_aws_lambda = "AWS_Lambda_" in aws_env

if env_is_test:
    from .agent.test import TestAgent
    from .recorder import StanRecorder

    agent = TestAgent()
    span_recorder = StanRecorder(agent)
    profiler = Profiler(agent)

elif env_is_aws_lambda:
    from .agent.aws_lambda import AWSLambdaAgent
    from .recorder import StanRecorder

    agent = AWSLambdaAgent()
    span_recorder = StanRecorder(agent)

elif env_is_aws_fargate:
    from .agent.aws_fargate import AWSFargateAgent
    from .recorder import StanRecorder

    agent = AWSFargateAgent()
    span_recorder = StanRecorder(agent)

else:
    from .agent.host import HostAgent
    from .recorder import StanRecorder

    agent = HostAgent()
    span_recorder = StanRecorder(agent)
    profiler = Profiler(agent)


def get_agent():
    """
    Retrieve the globally configured agent
    @return: The Instana Agent singleton
    """
    global agent
    return agent


def set_agent(new_agent):
    """
    Set the global agent for the Instana package.  This is used for the
    test suite only currently.

    @param new_agent: agent to replace current singleton
    @return: None
    """
    global agent
    agent = new_agent


# The global OpenTracing compatible tracer used internally by
# this package.
tracer = InstanaTracer(recorder=span_recorder)

if sys.version_info >= (3, 4):
    try:
        from opentracing.scope_managers.asyncio import AsyncioScopeManager
        async_tracer = InstanaTracer(scope_manager=AsyncioScopeManager(), recorder=span_recorder)
    except Exception:
        logger.debug("Error setting up async_tracer:", exc_info=True)


# Mock the tornado tracer until tornado is detected and instrumented first
tornado_tracer = tracer


def setup_tornado_tracer():
    global tornado_tracer
    from opentracing.scope_managers.tornado import TornadoScopeManager
    tornado_tracer = InstanaTracer(scope_manager=TornadoScopeManager(), recorder=span_recorder)


# Set ourselves as the tracer.
opentracing.tracer = tracer


def get_tracer():
    """
    Retrieve the globally configured tracer
    @return: Tracer
    """
    global tracer
    return tracer


def set_tracer(new_tracer):
    """
    Set the global tracer for the Instana package.  This is used for the
    test suite only currently.
    @param new_tracer: The new tracer to replace the singleton
    @return: None
    """
    global tracer
    tracer = new_tracer

def get_profiler():
    """
    Retrieve the globally configured profiler
    @return: Profiler
    """
    global profiler
    return profiler


def set_profiler(new_profiler):
    """
    Set the global profiler for the Instana package.  This is used for the
    test suite only currently.
    @param new_profiler: The new profiler to replace the singleton
    @return: None
    """
    global profiler
    profiler = new_profiler



from __future__ import absolute_import

import os
import pytest
import gevent
from gevent.pool import Group
import urllib3
import unittest

import tests.apps.flask_app
from instana.span import SDKSpan
from instana.singletons import tracer
from ..helpers import testenv, get_spans_by_filter
from opentracing.scope_managers.gevent import GeventScopeManager


@pytest.mark.skipif("GEVENT_TEST" not in os.environ, reason="")
class TestGEvent(unittest.TestCase):
    def setUp(self):
        self.http = urllib3.HTTPConnectionPool('127.0.0.1', port=testenv["wsgi_port"], maxsize=20)
        self.recorder = tracer.recorder
        self.recorder.clear_spans()
        tracer._scope_manager = GeventScopeManager()

    def tearDown(self):
        """ Do nothing for now """
        pass

    def make_http_call(self, n=None):
        return self.http.request('GET', testenv["wsgi_server"] + '/')

    def spawn_calls(self):
        with tracer.start_active_span('spawn_calls'):
            jobs = []
            jobs.append(gevent.spawn(self.make_http_call))
            jobs.append(gevent.spawn(self.make_http_call))
            jobs.append(gevent.spawn(self.make_http_call))
            gevent.joinall(jobs, timeout=2)

    def spawn_imap_unordered(self):
        igroup = Group()
        result = []
        with tracer.start_active_span('test'):
            for i in igroup.imap_unordered(self.make_http_call, range(3)):
                result.append(i)

    def launch_gevent_chain(self):
        with tracer.start_active_span('test'):
            gevent.spawn(self.spawn_calls).join()

    def test_spawning(self):
        gevent.spawn(self.launch_gevent_chain)

        gevent.sleep(2)

        spans = self.recorder.queued_spans()

        self.assertEqual(8, len(spans))

        span_filter = lambda span: span.n == "sdk" \
                                   and span.data['sdk']['name'] == 'test' and span.p == None
        test_spans = get_spans_by_filter(spans, span_filter)
        self.assertIsNotNone(test_spans)
        self.assertEqual(len(test_spans), 1)

        test_span = test_spans[0]
        self.assertTrue(type(test_spans[0]) is SDKSpan)

        span_filter = lambda span: span.n == "sdk" \
                                   and span.data['sdk']['name'] == 'spawn_calls' and span.p == test_span.s
        spawn_spans = get_spans_by_filter(spans, span_filter)
        self.assertIsNotNone(spawn_spans)
        self.assertEqual(len(spawn_spans), 1)

        spawn_span = spawn_spans[0]
        self.assertTrue(type(spawn_spans[0]) is SDKSpan)

        span_filter = lambda span: span.n == "urllib3"
        urllib3_spans = get_spans_by_filter(spans, span_filter)

        for urllib3_span in urllib3_spans:
            # spans should all have the same test span parent
            self.assertEqual(urllib3_span.t, spawn_span.t)
            self.assertEqual(urllib3_span.p, spawn_span.s)

            # find the wsgi span generated from this urllib3 request
            span_filter = lambda span: span.n == "wsgi" and span.p == urllib3_span.s
            wsgi_spans = get_spans_by_filter(spans, span_filter)
            self.assertIsNotNone(wsgi_spans)
            self.assertEqual(len(wsgi_spans), 1)

    def test_imap_unordered(self):
        gevent.spawn(self.spawn_imap_unordered())

        gevent.sleep(2)

        spans = self.recorder.queued_spans()
        self.assertEqual(7, len(spans))

        span_filter = lambda span: span.n == "sdk" \
                                   and span.data['sdk']['name'] == 'test' and span.p == None
        test_spans = get_spans_by_filter(spans, span_filter)
        self.assertIsNotNone(test_spans)
        self.assertEqual(len(test_spans), 1)

        test_span = test_spans[0]
        self.assertTrue(type(test_spans[0]) is SDKSpan)

        span_filter = lambda span: span.n == "urllib3"
        urllib3_spans = get_spans_by_filter(spans, span_filter)
        self.assertEqual(len(urllib3_spans), 3)

        for urllib3_span in urllib3_spans:
            # spans should all have the same test span parent
            self.assertEqual(urllib3_span.t, test_span.t)
            self.assertEqual(urllib3_span.p, test_span.s)

            # find the wsgi span generated from this urllib3 request
            span_filter = lambda span: span.n == "wsgi" and span.p == urllib3_span.s
            wsgi_spans = get_spans_by_filter(spans, span_filter)
            self.assertIsNotNone(wsgi_spans)
            self.assertEqual(len(wsgi_spans), 1)




from __future__ import absolute_import

import sys

from ..log import logger
from ..util import header_to_id
from ..span_context import SpanContext

PY2 = sys.version_info[0] == 2
PY3 = sys.version_info[0] == 3

# The carrier can be a dict or a list.
# Using the trace header as an example, it can be in the following forms
# for extraction:
#   X-Instana-T
#   HTTP_X_INSTANA_T
#
# The second form above is found in places like Django middleware for
# incoming requests.
#
# For injection, we only support the standard format:
#   X-Instana-T


class BasePropagator():
    UC_HEADER_KEY_T = 'X-INSTANA-T'
    UC_HEADER_KEY_S = 'X-INSTANA-S'
    UC_HEADER_KEY_L = 'X-INSTANA-L'
    UC_HEADER_KEY_SYNTHETIC = 'X-INSTANA-SYNTHETIC'

    HEADER_KEY_T = 'X-Instana-T'
    HEADER_KEY_S = 'X-Instana-S'
    HEADER_KEY_L = 'X-Instana-L'
    HEADER_KEY_SYNTHETIC = 'X-Instana-Synthetic'

    LC_HEADER_KEY_T = 'x-instana-t'
    LC_HEADER_KEY_S = 'x-instana-s'
    LC_HEADER_KEY_L = 'x-instana-l'
    LC_HEADER_KEY_SYNTHETIC = 'x-instana-synthetic'

    ALT_HEADER_KEY_T = 'HTTP_X_INSTANA_T'
    ALT_HEADER_KEY_S = 'HTTP_X_INSTANA_S'
    ALT_HEADER_KEY_L = 'HTTP_X_INSTANA_L'
    ALT_LC_HEADER_KEY_T = 'http_x_instana_t'
    ALT_LC_HEADER_KEY_S = 'http_x_instana_s'
    ALT_LC_HEADER_KEY_L = 'http_x_instana_l'
    ALT_LC_HEADER_KEY_SYNTHETIC = 'http_x_instana_synthetic'

    def extract(self, carrier):
        """
        Search carrier for the *HEADER* keys and return a SpanContext or None
        
        Note: Extract is on the base class since it never really varies in task regardless
        of the propagator in uses.

        :param carrier: The dict or list potentially containing context
        :return: SpanContext or None
        """
        trace_id = None
        span_id = None
        level = 1
        synthetic = False
        dc = None

        try:
            # Attempt to convert incoming <carrier> into a dict
            try:
                if isinstance(carrier, dict):
                    dc = carrier
                elif hasattr(carrier, "__dict__"):
                    dc = carrier.__dict__
                else:
                    dc = dict(carrier)
            except Exception:
                logger.debug("extract: Couln't convert %s", carrier)

            if dc is None:
                return None

            # Headers can exist in the standard X-Instana-T/S format or the alternate HTTP_X_INSTANA_T/S style
            # We do a case insensitive search to cover all possible variations of incoming headers.
            for key in dc.keys():
                lc_key = None

                if PY3 is True and isinstance(key, bytes):
                    lc_key = key.decode("utf-8").lower()
                else:
                    lc_key = key.lower()

                if self.LC_HEADER_KEY_T == lc_key:
                    trace_id = header_to_id(dc[key])
                elif self.LC_HEADER_KEY_S == lc_key:
                    span_id = header_to_id(dc[key])
                elif self.LC_HEADER_KEY_L == lc_key:
                    level = dc[key]
                elif self.LC_HEADER_KEY_SYNTHETIC == lc_key:
                    synthetic = dc[key] in ['1', b'1']

                elif self.ALT_LC_HEADER_KEY_T == lc_key:
                    trace_id = header_to_id(dc[key])
                elif self.ALT_LC_HEADER_KEY_S == lc_key:
                    span_id = header_to_id(dc[key])
                elif self.ALT_LC_HEADER_KEY_L == lc_key:
                    level = dc[key]
                elif self.ALT_LC_HEADER_KEY_SYNTHETIC == lc_key:
                    synthetic = dc[key] in ['1', b'1']

            ctx = None
            if trace_id is not None and span_id is not None:
                ctx = SpanContext(span_id=span_id,
                                  trace_id=trace_id,
                                  level=level,
                                  baggage={},
                                  sampled=True,
                                  synthetic=synthetic)
            elif synthetic:
                ctx = SpanContext(synthetic=synthetic)

            return ctx

        except Exception:
            logger.debug("extract error:", exc_info=True)


"""
Instrumentation for the gevent package.
"""
from __future__ import absolute_import

import sys
from ..log import logger
from ..singletons import tracer


def instrument_gevent():
    """ Adds context propagation to gevent greenlet spawning """
    try:
        logger.debug("Instrumenting gevent")

        import gevent
        from opentracing.scope_managers.gevent import GeventScopeManager
        from opentracing.scope_managers.gevent import _GeventScope

        def spawn_callback(new_greenlet):
            """ Handles context propagation for newly spawning greenlets """
            parent_scope = tracer.scope_manager.active
            if parent_scope is not None:
                # New greenlet, new clean slate.  Clone and make active in this new greenlet
                # the currently active scope (but don't finish() the span on close - it's a
                # clone/not the original and we don't want to close it prematurely)
                # TODO: Change to our own ScopeManagers
                parent_scope_clone = _GeventScope(parent_scope.manager, parent_scope.span, finish_on_close=False)
                tracer._scope_manager._set_greenlet_scope(parent_scope_clone, new_greenlet)

        logger.debug(" -> Updating tracer to use gevent based context management")
        tracer._scope_manager = GeventScopeManager()
        gevent.Greenlet.add_spawn_callback(spawn_callback)
    except:
        logger.debug("instrument_gevent: ", exc_info=True)


if 'gevent' in sys.modules:
    if sys.modules['gevent'].version_info < (1, 4):
        logger.debug("gevent < 1.4 detected.  The Instana package supports gevent versions 1.4 and greater.")
    else:
        instrument_gevent()
else:
    logger.debug("Instrumenting gevent: gevent not detected or loaded.  Nothing done.")




import threading
import os
import re
import importlib

from .runtime import runtime_info

class FrameCache(object):
    MAX_CACHE_SIZE = 2500

    def __init__(self, profiler):
        self.profiler = profiler
        self.profiler_frame_cache = None

        self.include_profiler_frames = None

        self.profiler_dir = os.path.dirname(os.path.realpath(__file__))

    def start(self):
        self.profiler_frame_cache = dict()

        self.include_profiler_frames = self.profiler.get_option('include_profiler_frames', False)

    def stop(self):
        pass

    def is_profiler_frame(self, filename):
        if filename in self.profiler_frame_cache:
            return self.profiler_frame_cache[filename]

        profiler_frame = False

        if not self.include_profiler_frames:
            if filename.startswith(self.profiler_dir):
                profiler_frame = True

        if len(self.profiler_frame_cache) < self.MAX_CACHE_SIZE:
            self.profiler_frame_cache[filename] = profiler_frame

        return profiler_frame



""" Collection helper for the Python runtime """
import os
import gc
import sys
import platform
import resource
import threading
from types import ModuleType
from pkg_resources import DistributionNotFound, get_distribution

from instana.log import logger
from instana.version import VERSION
from instana.util import DictionaryOfStan, determine_service_name

from .base import BaseHelper


class RuntimeHelper(BaseHelper):
    """ Helper class to collect snapshot and metrics for this Python runtime """
    def __init__(self, collector):
        super(RuntimeHelper, self).__init__(collector)
        self.previous = DictionaryOfStan()
        self.previous_rusage = resource.getrusage(resource.RUSAGE_SELF)

        if gc.isenabled():
            self.previous_gc_count = gc.get_count()
        else:
            self.previous_gc_count = None

    def collect_metrics(self, with_snapshot=False):
        plugin_data = dict()
        try:
            plugin_data["name"] = "com.instana.plugin.python"
            plugin_data["entityId"] = str(os.getpid())
            plugin_data["data"] = DictionaryOfStan()

            if hasattr(self.collector.agent, "announce_data"):
                try:
                    plugin_data["data"]["pid"] = self.collector.agent.announce_data.pid
                except Exception:
                    plugin_data["data"]["pid"] = str(os.getpid())
            else:
                plugin_data["data"]["pid"] = str(os.getpid())

            self._collect_runtime_metrics(plugin_data, with_snapshot)

            if with_snapshot is True:
                self._collect_runtime_snapshot(plugin_data)
        except Exception:
            logger.debug("_collect_metrics: ", exc_info=True)
        return [plugin_data]

    def _collect_runtime_metrics(self, plugin_data, with_snapshot):
        """ Collect up and return the runtime metrics """
        try:
            rusage = resource.getrusage(resource.RUSAGE_SELF)
            if gc.isenabled():
                self._collect_gc_metrics(plugin_data, with_snapshot)

            self._collect_thread_metrics(plugin_data, with_snapshot)

            value_diff = rusage.ru_utime - self.previous_rusage.ru_utime
            self.apply_delta(value_diff, self.previous['data']['metrics'],
                             plugin_data['data']['metrics'], "ru_utime", with_snapshot)

            value_diff = rusage.ru_stime - self.previous_rusage.ru_stime
            self.apply_delta(value_diff, self.previous['data']['metrics'],
                             plugin_data['data']['metrics'], "ru_stime", with_snapshot)

            self.apply_delta(rusage.ru_maxrss, self.previous['data']['metrics'],
                             plugin_data['data']['metrics'], "ru_maxrss", with_snapshot)
            self.apply_delta(rusage.ru_ixrss, self.previous['data']['metrics'],
                             plugin_data['data']['metrics'], "ru_ixrss", with_snapshot)
            self.apply_delta(rusage.ru_idrss, self.previous['data']['metrics'],
                             plugin_data['data']['metrics'], "ru_idrss", with_snapshot)
            self.apply_delta(rusage.ru_isrss, self.previous['data']['metrics'],
                             plugin_data['data']['metrics'], "ru_isrss", with_snapshot)

            value_diff = rusage.ru_minflt - self.previous_rusage.ru_minflt
            self.apply_delta(value_diff, self.previous['data']['metrics'],
                             plugin_data['data']['metrics'], "ru_minflt", with_snapshot)

            value_diff = rusage.ru_majflt - self.previous_rusage.ru_majflt
            self.apply_delta(value_diff, self.previous['data']['metrics'],
                             plugin_data['data']['metrics'], "ru_majflt", with_snapshot)

            value_diff = rusage.ru_nswap - self.previous_rusage.ru_nswap
            self.apply_delta(value_diff, self.previous['data']['metrics'],
                             plugin_data['data']['metrics'], "ru_nswap", with_snapshot)

            value_diff = rusage.ru_inblock - self.previous_rusage.ru_inblock
            self.apply_delta(value_diff, self.previous['data']['metrics'],
                             plugin_data['data']['metrics'], "ru_inblock", with_snapshot)

            value_diff = rusage.ru_oublock - self.previous_rusage.ru_oublock
            self.apply_delta(value_diff, self.previous['data']['metrics'],
                             plugin_data['data']['metrics'], "ru_oublock", with_snapshot)

            value_diff = rusage.ru_msgsnd - self.previous_rusage.ru_msgsnd
            self.apply_delta(value_diff, self.previous['data']['metrics'],
                             plugin_data['data']['metrics'], "ru_msgsnd", with_snapshot)

            value_diff = rusage.ru_msgrcv - self.previous_rusage.ru_msgrcv
            self.apply_delta(value_diff, self.previous['data']['metrics'],
                             plugin_data['data']['metrics'], "ru_msgrcv", with_snapshot)

            value_diff = rusage.ru_nsignals - self.previous_rusage.ru_nsignals
            self.apply_delta(value_diff, self.previous['data']['metrics'],
                             plugin_data['data']['metrics'], "ru_nsignals", with_snapshot)

            value_diff = rusage.ru_nvcsw - self.previous_rusage.ru_nvcsw
            self.apply_delta(value_diff, self.previous['data']['metrics'],
                             plugin_data['data']['metrics'], "ru_nvcsw", with_snapshot)

            value_diff = rusage.ru_nivcsw - self.previous_rusage.ru_nivcsw
            self.apply_delta(value_diff, self.previous['data']['metrics'],
                             plugin_data['data']['metrics'], "ru_nivcsw", with_snapshot)
        except Exception:
            logger.debug("_collect_runtime_metrics", exc_info=True)
        finally:
            self.previous_rusage = rusage

    def _collect_gc_metrics(self, plugin_data, with_snapshot):
        try:
            gc_count = gc.get_count()
            gc_threshold = gc.get_threshold()

            self.apply_delta(gc_count[0], self.previous['data']['metrics']['gc'],
                             plugin_data['data']['metrics']['gc'], "collect0", with_snapshot)
            self.apply_delta(gc_count[1], self.previous['data']['metrics']['gc'],
                             plugin_data['data']['metrics']['gc'], "collect1", with_snapshot)
            self.apply_delta(gc_count[2], self.previous['data']['metrics']['gc'],
                             plugin_data['data']['metrics']['gc'], "collect2", with_snapshot)

            self.apply_delta(gc_threshold[0], self.previous['data']['metrics']['gc'],
                             plugin_data['data']['metrics']['gc'], "threshold0", with_snapshot)
            self.apply_delta(gc_threshold[1], self.previous['data']['metrics']['gc'],
                             plugin_data['data']['metrics']['gc'], "threshold1", with_snapshot)
            self.apply_delta(gc_threshold[2], self.previous['data']['metrics']['gc'],
                             plugin_data['data']['metrics']['gc'], "threshold2", with_snapshot)
        except Exception:
            logger.debug("_collect_gc_metrics", exc_info=True)

    def _collect_thread_metrics(self, plugin_data, with_snapshot):
        try:
            threads = threading.enumerate()
            daemon_threads = [thread.daemon is True for thread in threads].count(True)
            self.apply_delta(daemon_threads, self.previous['data']['metrics'],
                             plugin_data['data']['metrics'], "daemon_threads", with_snapshot)

            alive_threads = [thread.daemon is False for thread in threads].count(True)
            self.apply_delta(alive_threads, self.previous['data']['metrics'],
                             plugin_data['data']['metrics'], "alive_threads", with_snapshot)

            dummy_threads = [isinstance(thread, threading._DummyThread) for thread in threads].count(True) # pylint: disable=protected-access
            self.apply_delta(dummy_threads, self.previous['data']['metrics'],
                             plugin_data['data']['metrics'], "dummy_threads", with_snapshot)
        except Exception:
            logger.debug("_collect_thread_metrics", exc_info=True)

    def _collect_runtime_snapshot(self,plugin_data):
        """ Gathers Python specific Snapshot information for this process """
        snapshot_payload = {}
        try:
            snapshot_payload['name'] = determine_service_name()
            snapshot_payload['version'] = sys.version
            snapshot_payload['f'] = platform.python_implementation() # flavor
            snapshot_payload['a'] = platform.architecture()[0] # architecture
            snapshot_payload['versions'] = self.gather_python_packages()
            snapshot_payload['iv'] = VERSION

            if 'AUTOWRAPT_BOOTSTRAP' in os.environ:
                snapshot_payload['m'] = 'Autowrapt'
            elif 'INSTANA_MAGIC' in os.environ:
                snapshot_payload['m'] = 'AutoTrace'
            else:
                snapshot_payload['m'] = 'Manual'

            try:
                from django.conf import settings # pylint: disable=import-outside-toplevel
                if hasattr(settings, 'MIDDLEWARE') and settings.MIDDLEWARE is not None:
                    snapshot_payload['djmw'] = settings.MIDDLEWARE
                elif hasattr(settings, 'MIDDLEWARE_CLASSES') and settings.MIDDLEWARE_CLASSES is not None:
                    snapshot_payload['djmw'] = settings.MIDDLEWARE_CLASSES
            except Exception:
                pass
        except Exception:
            logger.debug("collect_snapshot: ", exc_info=True)

        plugin_data['data']['snapshot'] = snapshot_payload

    def gather_python_packages(self):
        """ Collect up the list of modules in use """
        versions = dict()
        try:
            sys_packages = sys.modules.copy()

            for pkg_name in sys_packages:
                # Don't report submodules (e.g. django.x, django.y, django.z)
                # Skip modules that begin with underscore
                if ('.' in pkg_name) or pkg_name[0] == '_':
                    continue

                # Skip builtins
                if pkg_name in ["sys", "curses"]:
                    continue

                if sys_packages[pkg_name]:
                    try:
                        pkg_info = sys_packages[pkg_name].__dict__
                        if "__version__" in pkg_info:
                            if isinstance(pkg_info["__version__"], str):
                                versions[pkg_name] = pkg_info["__version__"]
                            else:
                                versions[pkg_name] = self.jsonable(pkg_info["__version__"])
                        elif "version" in pkg_info:
                            versions[pkg_name] = self.jsonable(pkg_info["version"])
                        else:
                            versions[pkg_name] = get_distribution(pkg_name).version
                    except DistributionNotFound:
                        pass
                    except Exception:
                        logger.debug("gather_python_packages: could not process module: %s", pkg_name)
            
            # Manually set our package version
            versions['instana'] = VERSION
        except Exception:
            logger.debug("gather_python_packages", exc_info=True)

        return versions

    def jsonable(self, value):
        try:
            if callable(value):
                try:
                    result = value()
                except Exception:
                    result = 'Unknown'
            elif isinstance(value, ModuleType):
                result = value
            else:
                result = value
            return str(result)
        except Exception:
            logger.debug("jsonable: ", exc_info=True)



import math
import os
import uuid
import time


class Profile(object):
    CATEGORY_CPU = 'cpu'
    CATEGORY_MEMORY = 'memory'
    CATEGORY_TIME = 'time'
    TYPE_CPU_USAGE = 'cpu-usage'
    TYPE_MEMORY_ALLOCATION_RATE = 'memory-allocation-rate'
    TYPE_BLOCKING_CALLS = 'blocking-calls'
    UNIT_NONE = ''
    UNIT_MILLISECOND = 'millisecond'
    UNIT_MICROSECOND = 'microsecond'
    UNIT_NANOSECOND = 'nanosecond'
    UNIT_BYTE = 'byte'
    UNIT_KILOBYTE = 'kilobyte'
    UNIT_PERCENT = 'percent'
    UNIT_SAMPLE = 'sample'
    RUNTIME_PYTHON = 'python'

    def __init__(self, category, typ, unit, roots, duration, timespan):
        self.process_id = str(os.getpid())
        self.id = generate_uuid()
        self.runtime = Profile.RUNTIME_PYTHON
        self.category = category
        self.type = typ
        self.unit = unit
        self.roots = roots
        self.duration = duration
        self.timespan = timespan
        self.timestamp = millis()

    def to_dict(self):
        profile_dict = {
            'pid': self.process_id,
            'id': self.id,
            'runtime': self.runtime,
            'category': self.category,
            'type': self.type,
            'unit': self.unit,
            'roots': [root.to_dict() for root in self.roots],
            'duration': self.duration,
            'timespan': self.timespan,
            'timestamp': self.timestamp
        }

        return profile_dict


class CallSite:
    __slots__ = [
        'method_name',
        'file_name',
        'file_line',
        'measurement',
        'num_samples',
        'children'
    ]

    def __init__(self, method_name, file_name, file_line):
        self.method_name = method_name
        self.file_name = file_name
        self.file_line = file_line
        self.measurement = 0
        self.num_samples = 0
        self.children = dict()

    def create_key(self, method_name, file_name, file_line):
        return '{0} ({1}:{2})'.format(method_name, file_name, file_line)

    def find_child(self, method_name, file_name, file_line):
        key = self.create_key(method_name, file_name, file_line)
        if key in self.children:
            return self.children[key]

        return None

    def add_child(self, child):
        self.children[self.create_key(child.method_name, child.file_name, child.file_line)] = child

    def remove_child(self, child):
        del self.children[self.create_key(child.method_name, child.file_name, child.file_line)]

    def find_or_add_child(self, method_name, file_name, file_line):
        child = self.find_child(method_name, file_name, file_line)
        if child == None:
            child = CallSite(method_name, file_name, file_line)
            self.add_child(child)

        return child

    def increment(self, value, count):
        self.measurement += value
        self.num_samples += count

    def normalize(self, factor):
        self.measurement = self.measurement / factor
        self.num_samples = int(math.ceil(self.num_samples / factor))

        for child in self.children.values():
            child.normalize(factor)

    def floor(self):
        self.measurement = int(self.measurement)

        for child in self.children.values():
            child.floor()

    def to_dict(self):
        children_dicts = []
        for child in self.children.values():
            children_dicts.append(child.to_dict())

        call_site_dict = {
            'method_name': self.method_name,
            'file_name': self.file_name,
            'file_line': self.file_line,
            'measurement': self.measurement,
            'num_samples': self.num_samples,
            'children': children_dicts
        }

        return call_site_dict


def millis():
    return int(round(time.time() * 1000))


def generate_uuid():
    return str(uuid.uuid4())



from __future__ import absolute_import

import time
import pytest
import requests
import multiprocessing
from ..helpers import testenv
from instana.singletons import tracer
from ..helpers import get_first_span_by_filter

@pytest.fixture(scope="module")
def server():
    from tests.apps.starlette_app import launch_starlette
    proc = multiprocessing.Process(target=launch_starlette, args=(), daemon=True)
    proc.start()
    time.sleep(2)
    yield
    proc.kill() # Kill server after tests

def test_vanilla_get(server):
    result = requests.get(testenv["starlette_server"] + '/')
    assert(result)
    spans = tracer.recorder.queued_spans()
    # Starlette instrumentation (like all instrumentation) _always_ traces unless told otherwise
    assert len(spans) == 1
    assert spans[0].n == 'asgi'

    assert "X-Instana-T" in result.headers
    assert "X-Instana-S" in result.headers
    assert "X-Instana-L" in result.headers
    assert result.headers["X-Instana-L"] == '1'
    assert "Server-Timing" in result.headers

def test_basic_get(server):
    result = None
    with tracer.start_active_span('test'):
        result = requests.get(testenv["starlette_server"] + '/')

    assert(result)

    spans = tracer.recorder.queued_spans()
    assert len(spans) == 3

    span_filter = lambda span: span.n == "sdk"
    test_span = get_first_span_by_filter(spans, span_filter)
    assert(test_span)

    span_filter = lambda span: span.n == "urllib3"
    urllib3_span = get_first_span_by_filter(spans, span_filter)
    assert(urllib3_span)

    span_filter = lambda span: span.n == "asgi"
    asgi_span = get_first_span_by_filter(spans, span_filter)
    assert(asgi_span)

    assert(test_span.t == urllib3_span.t == asgi_span.t)
    assert(asgi_span.p == urllib3_span.s)
    assert(urllib3_span.p == test_span.s)

    assert "X-Instana-T" in result.headers
    assert result.headers["X-Instana-T"] == asgi_span.t
    assert "X-Instana-S" in result.headers
    assert result.headers["X-Instana-S"] == asgi_span.s
    assert "X-Instana-L" in result.headers
    assert result.headers["X-Instana-L"] == '1'
    assert "Server-Timing" in result.headers
    assert result.headers["Server-Timing"] == ("intid;desc=%s" % asgi_span.t)

    assert('http' in asgi_span.data)
    assert(asgi_span.ec == None)
    assert(isinstance(asgi_span.stack, list))
    assert(asgi_span.data['http']['host'] == '127.0.0.1')
    assert(asgi_span.data['http']['path'] == '/')
    assert(asgi_span.data['http']['path_tpl'] == '/')
    assert(asgi_span.data['http']['method'] == 'GET')
    assert(asgi_span.data['http']['status'] == 200)
    assert(asgi_span.data['http']['error'] == None)

def test_path_templates(server):
    result = None
    with tracer.start_active_span('test'):
        result = requests.get(testenv["starlette_server"] + '/users/1')

    assert(result)

    spans = tracer.recorder.queued_spans()
    assert len(spans) == 3

    span_filter = lambda span: span.n == "sdk"
    test_span = get_first_span_by_filter(spans, span_filter)
    assert(test_span)

    span_filter = lambda span: span.n == "urllib3"
    urllib3_span = get_first_span_by_filter(spans, span_filter)
    assert(urllib3_span)

    span_filter = lambda span: span.n == "asgi"
    asgi_span = get_first_span_by_filter(spans, span_filter)
    assert(asgi_span)

    assert(test_span.t == urllib3_span.t == asgi_span.t)
    assert(asgi_span.p == urllib3_span.s)
    assert(urllib3_span.p == test_span.s)

    assert "X-Instana-T" in result.headers
    assert result.headers["X-Instana-T"] == asgi_span.t
    assert "X-Instana-S" in result.headers
    assert result.headers["X-Instana-S"] == asgi_span.s
    assert "X-Instana-L" in result.headers
    assert result.headers["X-Instana-L"] == '1'
    assert "Server-Timing" in result.headers
    assert result.headers["Server-Timing"] == ("intid;desc=%s" % asgi_span.t)

    assert('http' in asgi_span.data)
    assert(asgi_span.ec == None)
    assert(isinstance(asgi_span.stack, list))
    assert(asgi_span.data['http']['host'] == '127.0.0.1')
    assert(asgi_span.data['http']['path'] == '/users/1')
    assert(asgi_span.data['http']['path_tpl'] == '/users/{user_id}')
    assert(asgi_span.data['http']['params'] == None)
    assert(asgi_span.data['http']['method'] == 'GET')
    assert(asgi_span.data['http']['status'] == 200)
    assert(asgi_span.data['http']['error'] == None)

def test_secret_scrubbing(server):
    result = None
    with tracer.start_active_span('test'):
        result = requests.get(testenv["starlette_server"] + '/?secret=shhh')

    assert(result)

    spans = tracer.recorder.queued_spans()
    assert len(spans) == 3

    span_filter = lambda span: span.n == "sdk"
    test_span = get_first_span_by_filter(spans, span_filter)
    assert(test_span)

    span_filter = lambda span: span.n == "urllib3"
    urllib3_span = get_first_span_by_filter(spans, span_filter)
    assert(urllib3_span)

    span_filter = lambda span: span.n == "asgi"
    asgi_span = get_first_span_by_filter(spans, span_filter)
    assert(asgi_span)

    assert(test_span.t == urllib3_span.t == asgi_span.t)
    assert(asgi_span.p == urllib3_span.s)
    assert(urllib3_span.p == test_span.s)

    assert "X-Instana-T" in result.headers
    assert result.headers["X-Instana-T"] == asgi_span.t
    assert "X-Instana-S" in result.headers
    assert result.headers["X-Instana-S"] == asgi_span.s
    assert "X-Instana-L" in result.headers
    assert result.headers["X-Instana-L"] == '1'
    assert "Server-Timing" in result.headers
    assert result.headers["Server-Timing"] == ("intid;desc=%s" % asgi_span.t)

    assert('http' in asgi_span.data)
    assert(asgi_span.ec == None)
    assert(isinstance(asgi_span.stack, list))
    assert(asgi_span.data['http']['host'] == '127.0.0.1')
    assert(asgi_span.data['http']['path'] == '/')
    assert(asgi_span.data['http']['path_tpl'] == '/')
    assert(asgi_span.data['http']['params'] == 'secret=<redacted>')
    assert(asgi_span.data['http']['method'] == 'GET')
    assert(asgi_span.data['http']['status'] == 200)
    assert(asgi_span.data['http']['error'] == None)

def test_synthetic_request(server):
    request_headers = {
        'X-Instana-Synthetic': '1'
    }
    with tracer.start_active_span('test'):
        result = requests.get(testenv["starlette_server"] + '/', headers=request_headers)

    assert(result)

    spans = tracer.recorder.queued_spans()
    assert len(spans) == 3

    span_filter = lambda span: span.n == "sdk"
    test_span = get_first_span_by_filter(spans, span_filter)
    assert(test_span)

    span_filter = lambda span: span.n == "urllib3"
    urllib3_span = get_first_span_by_filter(spans, span_filter)
    assert(urllib3_span)

    span_filter = lambda span: span.n == "asgi"
    asgi_span = get_first_span_by_filter(spans, span_filter)
    assert(asgi_span)

    assert(test_span.t == urllib3_span.t == asgi_span.t)
    assert(asgi_span.p == urllib3_span.s)
    assert(urllib3_span.p == test_span.s)

    assert "X-Instana-T" in result.headers
    assert result.headers["X-Instana-T"] == asgi_span.t
    assert "X-Instana-S" in result.headers
    assert result.headers["X-Instana-S"] == asgi_span.s
    assert "X-Instana-L" in result.headers
    assert result.headers["X-Instana-L"] == '1'
    assert "Server-Timing" in result.headers
    assert result.headers["Server-Timing"] == ("intid;desc=%s" % asgi_span.t)

    assert('http' in asgi_span.data)
    assert(asgi_span.ec == None)
    assert(isinstance(asgi_span.stack, list))
    assert(asgi_span.data['http']['host'] == '127.0.0.1')
    assert(asgi_span.data['http']['path'] == '/')
    assert(asgi_span.data['http']['path_tpl'] == '/')
    assert(asgi_span.data['http']['method'] == 'GET')
    assert(asgi_span.data['http']['status'] == 200)
    assert(asgi_span.data['http']['error'] == None)

    assert(asgi_span.sy)
    assert(urllib3_span.sy is None)
    assert(test_span.sy is None)

def test_custom_header_capture(server):
    from instana.singletons import agent

    # The background Starlette server is pre-configured with custom headers to capture

    request_headers = {
        'X-Capture-This': 'this',
        'X-Capture-That': 'that'
    }
    with tracer.start_active_span('test'):
        result = requests.get(testenv["starlette_server"] + '/', headers=request_headers)

    assert(result)

    spans = tracer.recorder.queued_spans()
    assert len(spans) == 3

    span_filter = lambda span: span.n == "sdk"
    test_span = get_first_span_by_filter(spans, span_filter)
    assert(test_span)

    span_filter = lambda span: span.n == "urllib3"
    urllib3_span = get_first_span_by_filter(spans, span_filter)
    assert(urllib3_span)

    span_filter = lambda span: span.n == "asgi"
    asgi_span = get_first_span_by_filter(spans, span_filter)
    assert(asgi_span)

    assert(test_span.t == urllib3_span.t == asgi_span.t)
    assert(asgi_span.p == urllib3_span.s)
    assert(urllib3_span.p == test_span.s)

    assert "X-Instana-T" in result.headers
    assert result.headers["X-Instana-T"] == asgi_span.t
    assert "X-Instana-S" in result.headers
    assert result.headers["X-Instana-S"] == asgi_span.s
    assert "X-Instana-L" in result.headers
    assert result.headers["X-Instana-L"] == '1'
    assert "Server-Timing" in result.headers
    assert result.headers["Server-Timing"] == ("intid;desc=%s" % asgi_span.t)

    assert('http' in asgi_span.data)
    assert(asgi_span.ec == None)
    assert(isinstance(asgi_span.stack, list))
    assert(asgi_span.data['http']['host'] == '127.0.0.1')
    assert(asgi_span.data['http']['path'] == '/')
    assert(asgi_span.data['http']['path_tpl'] == '/')
    assert(asgi_span.data['http']['method'] == 'GET')
    assert(asgi_span.data['http']['status'] == 200)
    assert(asgi_span.data['http']['error'] == None)

    assert("http.X-Capture-This" in asgi_span.data["custom"]['tags'])
    assert("this" == asgi_span.data["custom"]['tags']["http.X-Capture-This"])
    assert("http.X-Capture-That" in asgi_span.data["custom"]['tags'])
    assert("that" == asgi_span.data["custom"]['tags']["http.X-Capture-That"])



from __future__ import absolute_import

import logging
import unittest
from ..helpers import testenv
from instana.singletons import tracer

import psycopg2
import psycopg2.extras
import psycopg2.extensions as ext

logger = logging.getLogger(__name__)

create_table_query = """
CREATE TABLE IF NOT EXISTS users(
   id serial PRIMARY KEY,
   name VARCHAR (50),
   password VARCHAR (50),
   email VARCHAR (355),
   created_on TIMESTAMP,
   last_login TIMESTAMP
);
"""

create_proc_query = """\
CREATE OR REPLACE FUNCTION test_proc(candidate VARCHAR(70)) 
RETURNS text AS $$
BEGIN
    RETURN(SELECT name FROM users where email = candidate);
END;
$$ LANGUAGE plpgsql;
"""

drop_proc_query = "DROP FUNCTION IF EXISTS test_proc(VARCHAR(70));"

db = psycopg2.connect(host=testenv['postgresql_host'], port=testenv['postgresql_port'],
                     user=testenv['postgresql_user'], password=testenv['postgresql_pw'],
                     database=testenv['postgresql_db'])

cursor = db.cursor()
cursor.execute(create_table_query)
cursor.execute(drop_proc_query)
cursor.execute(create_proc_query)
db.commit()
cursor.close()
db.close()


class TestPsycoPG2(unittest.TestCase):
    def setUp(self):
        self.db = psycopg2.connect(host=testenv['postgresql_host'], port=testenv['postgresql_port'],
                                   user=testenv['postgresql_user'], password=testenv['postgresql_pw'],
                                   database=testenv['postgresql_db'])
        self.cursor = self.db.cursor()
        self.recorder = tracer.recorder
        self.recorder.clear_spans()
        tracer.cur_ctx = None

    def tearDown(self):
        """ Do nothing for now """
        return None

    def test_vanilla_query(self):
        assert psycopg2.extras.register_uuid(None, self.db)
        assert psycopg2.extras.register_uuid(None, self.db.cursor())

        self.cursor.execute("""SELECT * from users""")
        result = self.cursor.fetchone()

        self.assertEqual(6, len(result))

        spans = self.recorder.queued_spans()
        self.assertEqual(0, len(spans))

    def test_basic_query(self):
        with tracer.start_active_span('test'):
            self.cursor.execute("""SELECT * from users""")
            self.cursor.fetchone()
            self.db.commit()

        spans = self.recorder.queued_spans()
        self.assertEqual(2, len(spans))

        db_span = spans[0]
        test_span = spans[1]

        self.assertEqual("test", test_span.data["sdk"]["name"])
        self.assertEqual(test_span.t, db_span.t)
        self.assertEqual(db_span.p, test_span.s)

        self.assertEqual(None, db_span.ec)

        self.assertEqual(db_span.n, "postgres")
        self.assertEqual(db_span.data["pg"]["db"], testenv['postgresql_db'])
        self.assertEqual(db_span.data["pg"]["user"], testenv['postgresql_user'])
        self.assertEqual(db_span.data["pg"]["stmt"], 'SELECT * from users')
        self.assertEqual(db_span.data["pg"]["host"], testenv['postgresql_host'])
        self.assertEqual(db_span.data["pg"]["port"], testenv['postgresql_port'])

    def test_basic_insert(self):
        with tracer.start_active_span('test'):
            self.cursor.execute("""INSERT INTO users(name, email) VALUES(%s, %s)""", ('beaker', 'beaker@muppets.com'))

        spans = self.recorder.queued_spans()
        self.assertEqual(2, len(spans))

        db_span = spans[0]
        test_span = spans[1]

        self.assertEqual("test", test_span.data["sdk"]["name"])
        self.assertEqual(test_span.t, db_span.t)
        self.assertEqual(db_span.p, test_span.s)

        self.assertEqual(None, db_span.ec)

        self.assertEqual(db_span.n, "postgres")
        self.assertEqual(db_span.data["pg"]["db"], testenv['postgresql_db'])
        self.assertEqual(db_span.data["pg"]["user"], testenv['postgresql_user'])
        self.assertEqual(db_span.data["pg"]["stmt"], 'INSERT INTO users(name, email) VALUES(%s, %s)')
        self.assertEqual(db_span.data["pg"]["host"], testenv['postgresql_host'])
        self.assertEqual(db_span.data["pg"]["port"], testenv['postgresql_port'])

    def test_executemany(self):
        result = None
        with tracer.start_active_span('test'):
            result = self.cursor.executemany("INSERT INTO users(name, email) VALUES(%s, %s)",
                                             [('beaker', 'beaker@muppets.com'), ('beaker', 'beaker@muppets.com')])
            self.db.commit()

        spans = self.recorder.queued_spans()
        self.assertEqual(2, len(spans))

        db_span = spans[0]
        test_span = spans[1]

        self.assertEqual("test", test_span.data["sdk"]["name"])
        self.assertEqual(test_span.t, db_span.t)
        self.assertEqual(db_span.p, test_span.s)

        self.assertEqual(None, db_span.ec)

        self.assertEqual(db_span.n, "postgres")
        self.assertEqual(db_span.data["pg"]["db"], testenv['postgresql_db'])
        self.assertEqual(db_span.data["pg"]["user"], testenv['postgresql_user'])
        self.assertEqual(db_span.data["pg"]["stmt"], 'INSERT INTO users(name, email) VALUES(%s, %s)')
        self.assertEqual(db_span.data["pg"]["host"], testenv['postgresql_host'])
        self.assertEqual(db_span.data["pg"]["port"], testenv['postgresql_port'])

    def test_call_proc(self):
        result = None
        with tracer.start_active_span('test'):
            result = self.cursor.callproc('test_proc', ('beaker',))

        assert(type(result) is tuple)

        spans = self.recorder.queued_spans()
        self.assertEqual(2, len(spans))

        db_span = spans[0]
        test_span = spans[1]

        self.assertEqual("test", test_span.data["sdk"]["name"])
        self.assertEqual(test_span.t, db_span.t)
        self.assertEqual(db_span.p, test_span.s)

        self.assertEqual(None, db_span.ec)

        self.assertEqual(db_span.n, "postgres")
        self.assertEqual(db_span.data["pg"]["db"], testenv['postgresql_db'])
        self.assertEqual(db_span.data["pg"]["user"], testenv['postgresql_user'])
        self.assertEqual(db_span.data["pg"]["stmt"], 'test_proc')
        self.assertEqual(db_span.data["pg"]["host"], testenv['postgresql_host'])
        self.assertEqual(db_span.data["pg"]["port"], testenv['postgresql_port'])

    def test_error_capture(self):
        result = None
        try:
            with tracer.start_active_span('test'):
                result = self.cursor.execute("""SELECT * from blah""")
                self.cursor.fetchone()
        except Exception:
            pass

        assert(result is None)

        spans = self.recorder.queued_spans()
        self.assertEqual(2, len(spans))

        db_span = spans[0]
        test_span = spans[1]

        self.assertEqual("test", test_span.data["sdk"]["name"])
        self.assertEqual(test_span.t, db_span.t)
        self.assertEqual(db_span.p, test_span.s)

        self.assertEqual(1, db_span.ec)
        self.assertEqual(db_span.data["pg"]["error"], 'relation "blah" does not exist\nLINE 1: SELECT * from blah\n                      ^\n')

        self.assertEqual(db_span.n, "postgres")
        self.assertEqual(db_span.data["pg"]["db"], testenv['postgresql_db'])
        self.assertEqual(db_span.data["pg"]["user"], testenv['postgresql_user'])
        self.assertEqual(db_span.data["pg"]["stmt"], 'SELECT * from blah')
        self.assertEqual(db_span.data["pg"]["host"], testenv['postgresql_host'])
        self.assertEqual(db_span.data["pg"]["port"], testenv['postgresql_port'])

    # Added to validate unicode support and register_type.
    def test_unicode(self):
        ext.register_type(ext.UNICODE, self.cursor)
        #
        # Python 2 chokes on Unicode and CircleCI tests are hanging (but pass locally).
        # Disable these tests for now as we want to really just test register_type
        # anyways
        #
        # snowman = "\u2603"
        #
        # self.cursor.execute("delete from users where id in (1,2,3)")
        #
        # # unicode in statement
        # psycopg2.extras.execute_batch(self.cursor,
        #     "insert into users (id, name) values (%%s, %%s) -- %s" % snowman, [(1, 'x')])
        # self.cursor.execute("select id, name from users where id = 1")
        # self.assertEqual(self.cursor.fetchone(), (1, 'x'))
        #
        # # unicode in data
        # psycopg2.extras.execute_batch(self.cursor,
        #     "insert into users (id, name) values (%s, %s)", [(2, snowman)])
        # self.cursor.execute("select id, name from users where id = 2")
        # self.assertEqual(self.cursor.fetchone(), (2, snowman))
        #
        # # unicode in both
        # psycopg2.extras.execute_batch(self.cursor,
        #     "insert into users (id, name) values (%%s, %%s) -- %s" % snowman, [(3, snowman)])
        # self.cursor.execute("select id, name from users where id = 3")
        # self.assertEqual(self.cursor.fetchone(), (3, snowman))

    def test_register_type(self):
        import uuid

        oid1 = 2950
        oid2 = 2951

        ext.UUID = ext.new_type((oid1,), "UUID", lambda data, cursor: data and uuid.UUID(data) or None)
        ext.UUIDARRAY = ext.new_array_type((oid2,), "UUID[]", ext.UUID)

        ext.register_type(ext.UUID, self.cursor)
        ext.register_type(ext.UUIDARRAY, self.cursor)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import logging
import opentracing.ext.tags as ext
from flask import jsonify, Response
from wsgiref.simple_server import make_server
from flask import Flask, redirect, render_template, render_template_string

try:
    import boto3
    from moto import mock_sqs
except ImportError:
    # Doesn't matter.  We won't call routes using boto3
    # in test sets that don't install/test for it.
    pass

from ...helpers import testenv
from instana.singletons import tracer

logging.basicConfig(level=logging.WARNING)
logger = logging.getLogger(__name__)

testenv["wsgi_port"] = 10811
testenv["wsgi_server"] = ("http://127.0.0.1:" + str(testenv["wsgi_port"]))

app = Flask(__name__)
app.debug = False
app.use_reloader = False

flask_server = make_server('127.0.0.1', testenv["wsgi_port"], app.wsgi_app)


class InvalidUsage(Exception):
    status_code = 400

    def __init__(self, message, status_code=None, payload=None):
        Exception.__init__(self)
        self.message = message
        if status_code is not None:
            self.status_code = status_code
        self.payload = payload

    def to_dict(self):
        rv = dict(self.payload or ())
        rv['message'] = self.message
        return rv


class NotFound(Exception):
    status_code = 404

    def __init__(self, message, status_code=None, payload=None):
        Exception.__init__(self)
        self.message = message
        if status_code is not None:
            self.status_code = status_code
        self.payload = payload

    def to_dict(self):
        rv = dict(self.payload or ())
        rv['message'] = self.message
        return rv


@app.route("/")
def hello():
    return "<center><h1>🐍 Hello Stan! 🦄</h1></center>"


@app.route("/users/<username>/sayhello")
def username_hello(username):
    return u"<center><h1>🐍 Hello %s! 🦄</h1></center>" % username


@app.route("/complex")
def gen_opentracing():
    with tracer.start_active_span('asteroid') as pscope:
        pscope.span.set_tag(ext.COMPONENT, "Python simple example app")
        pscope.span.set_tag(ext.SPAN_KIND, ext.SPAN_KIND_RPC_SERVER)
        pscope.span.set_tag(ext.PEER_HOSTNAME, "localhost")
        pscope.span.set_tag(ext.HTTP_URL, "/python/simple/one")
        pscope.span.set_tag(ext.HTTP_METHOD, "GET")
        pscope.span.set_tag(ext.HTTP_STATUS_CODE, 200)
        pscope.span.log_kv({"foo": "bar"})

        with tracer.start_active_span('spacedust', child_of=pscope.span) as cscope:
            cscope.span.set_tag(ext.SPAN_KIND, ext.SPAN_KIND_RPC_CLIENT)
            cscope.span.set_tag(ext.PEER_HOSTNAME, "localhost")
            cscope.span.set_tag(ext.HTTP_URL, "/python/simple/two")
            cscope.span.set_tag(ext.HTTP_METHOD, "POST")
            cscope.span.set_tag(ext.HTTP_STATUS_CODE, 204)
            cscope.span.set_baggage_item("someBaggage", "someValue")

    return "<center><h1>🐍 Generated some OT spans... 🦄</h1></center>"


@app.route("/301")
def threehundredone():
    return redirect('/', code=301)


@app.route("/302")
def threehundredtwo():
    return redirect('/', code=302)


@app.route("/400")
def fourhundred():
    return "Simulated Bad Request", 400


@app.route("/custom-404")
def custom404():
    raise NotFound("My custom 404 message")


@app.route("/405")
def fourhundredfive():
    return "Simulated Method not allowed", 405


@app.route("/500")
def fivehundred():
    return "Simulated Internal Server Error", 500


@app.route("/504")
def fivehundredfour():
    return "Simulated Gateway Timeout", 504


@app.route("/exception")
def exception():
    raise Exception('fake error')


@app.route("/exception-invalid-usage")
def exception_invalid_usage():
    raise InvalidUsage("Simulated custom exception", status_code=502)


@app.route("/render")
def render():
    return render_template('flask_render_template.html', name="Peter")


@app.route("/render_string")
def render_string():
    return render_template_string('hello {{ what }}', what='world')


@app.route("/render_error")
def render_error():
    return render_template('flask_render_error.html', what='world')


@app.route("/response_headers")
def response_headers():
    resp = Response("Foo bar baz")
    resp.headers['X-Capture-This'] = 'Ok'
    return resp


@app.route("/boto3/sqs")
def boto3_sqs():
    os.environ['AWS_ACCESS_KEY_ID'] = 'testing'
    os.environ['AWS_SECRET_ACCESS_KEY'] = 'testing'
    os.environ['AWS_SECURITY_TOKEN'] = 'testing'
    os.environ['AWS_SESSION_TOKEN'] = 'testing'

    with mock_sqs():
        boto3_client = boto3.client('sqs', region_name='us-east-1')
        response = boto3_client.create_queue(
            QueueName='SQS_QUEUE_NAME',
            Attributes={
                'DelaySeconds': '60',
                'MessageRetentionPeriod': '600'
            }
        )

        queue_url = response['QueueUrl']
        response = boto3_client.send_message(
                QueueUrl=queue_url,
                DelaySeconds=10,
                MessageAttributes={
                    'Website': {
                        'DataType': 'String',
                        'StringValue': 'https://www.instana.com'
                    },
                },
                MessageBody=('Monitor any application, service, or request '
                            'with Instana Application Performance Monitoring')
            )
        return Response(response)

@app.errorhandler(InvalidUsage)
def handle_invalid_usage(error):
    logger.error("InvalidUsage error handler invoked")
    response = jsonify(error.to_dict())
    response.status_code = error.status_code
    return response


@app.errorhandler(404)
@app.errorhandler(NotFound)
def handle_not_found(e):
    return "blah: %s" % str(e), 404


if __name__ == '__main__':
    flask_server.request_queue_size = 20
    flask_server.serve_forever()



import sys
import signal


class runtime_info(object):
    OS_LINUX = (sys.platform.startswith('linux'))
    OS_DARWIN = (sys.platform == 'darwin')
    OS_WIN = (sys.platform == 'win32')
    PYTHON_2 = (sys.version_info.major == 2)
    PYTHON_3 = (sys.version_info.major == 3)
    GEVENT = False

try:
    import gevent
    if hasattr(gevent, '_threading'):
        runtime_info.GEVENT = True
except ImportError:
    pass


def min_version(major, minor=0):
    return (sys.version_info.major == major and sys.version_info.minor >= minor)


def register_signal(signal_number, handler_func, once=False):
    prev_handler = None

    def _handler(signum, frame):
        skip_prev = handler_func(signum, frame)

        if not skip_prev:
            if callable(prev_handler):
                if once:
                    signal.signal(signum, prev_handler)
                prev_handler(signum, frame)
            elif prev_handler == signal.SIG_DFL and once:
                signal.signal(signum, signal.SIG_DFL)
                os.kill(os.getpid(), signum)

    prev_handler = signal.signal(signal_number, _handler)



from __future__ import absolute_import

from ..log import logger
from .base_propagator import BasePropagator


class BinaryPropagator(BasePropagator):
    """
    A Propagator for BINARY.
    The BINARY format represents SpanContexts in an opaque bytearray carrier.
    """
    
    # ByteArray variations from base class
    HEADER_KEY_T = b'x-instana-t'
    HEADER_KEY_S = b'x-instana-s'
    HEADER_KEY_L = b'x-instana-l'
    HEADER_SERVER_TIMING = b'server-timing'

    def inject(self, span_context, carrier):
        try:
            trace_id = str.encode(span_context.trace_id)
            span_id = str.encode(span_context.span_id)
            level = str.encode("1")
            server_timing = str.encode("intid;desc=%s" % span_context.trace_id)

            if isinstance(carrier, dict) or hasattr(carrier, "__dict__"):
                carrier[self.HEADER_KEY_T] = trace_id
                carrier[self.HEADER_KEY_S] = span_id
                carrier[self.HEADER_KEY_L] = level
                carrier[self.HEADER_SERVER_TIMING] = server_timing
            elif isinstance(carrier, list):
                carrier.append((self.HEADER_KEY_T, trace_id))
                carrier.append((self.HEADER_KEY_S, span_id))
                carrier.append((self.HEADER_KEY_L, level))
                carrier.append((self.HEADER_SERVER_TIMING, server_timing))
            elif isinstance(carrier, tuple):
                carrier = carrier.__add__(((self.HEADER_KEY_T, trace_id),))
                carrier = carrier.__add__(((self.HEADER_KEY_S, span_id),))
                carrier = carrier.__add__(((self.HEADER_KEY_L, level),))
                carrier = carrier.__add__(((self.HEADER_SERVER_TIMING, server_timing),))
            elif hasattr(carrier, '__setitem__'):
                carrier.__setitem__(self.HEADER_KEY_T, trace_id)
                carrier.__setitem__(self.HEADER_KEY_S, span_id)
                carrier.__setitem__(self.HEADER_KEY_L, level)
                carrier.__setitem__(self.HEADER_SERVER_TIMING, server_timing)
            else:
                raise Exception("Unsupported carrier type", type(carrier))

            return carrier
        except Exception:
            logger.debug("inject error:", exc_info=True)



from __future__ import absolute_import

from ..log import logger
from ..singletons import tracer

try:
    import pymongo
    from pymongo import monitoring
    from bson import json_util

    class MongoCommandTracer(monitoring.CommandListener):
        def __init__(self):
            self.__active_commands = {}

        def started(self, event):
            parent_span = tracer.active_span

            # return early if we're not tracing
            if parent_span is None:
                return

            with tracer.start_active_span("mongo", child_of=parent_span) as scope:
                self._collect_connection_tags(scope.span, event)
                self._collect_command_tags(scope.span, event)

                # include collection name into the namespace if provided
                if event.command_name in event.command:
                    scope.span.set_tag("collection", event.command.get(event.command_name))

                self.__active_commands[event.request_id] = scope

        def succeeded(self, event):
            active_span = self.__active_commands.pop(event.request_id, None)

            # return early if we're not tracing
            if active_span is None:
                return

        def failed(self, event):
            active_span = self.__active_commands.pop(event.request_id, None)

            # return early if we're not tracing
            if active_span is None:
                return

            active_span.log_exception(event.failure)

        def _collect_connection_tags(self, span, event):
            (host, port) = event.connection_id

            span.set_tag("host", host)
            span.set_tag("port", str(port))
            span.set_tag("db", event.database_name)

        def _collect_command_tags(self, span, event):
            """
            Extract MongoDB command name and arguments and attach it to the span
            """
            cmd = event.command_name
            span.set_tag("command", cmd)

            for key in ["filter", "query"]:
                if key in event.command:
                    span.set_tag("filter", json_util.dumps(event.command.get(key)))
                    break

            # The location of command documents within the command object depends on the name
            # of this command. This is the name -> command object key mapping
            cmd_doc_locations = {
                "insert": "documents",
                "update": "updates",
                "delete": "deletes",
                "aggregate": "pipeline"
            }

            cmd_doc = None
            if cmd in cmd_doc_locations:
                cmd_doc = event.command.get(cmd_doc_locations[cmd])
            elif cmd.lower() == "mapreduce": # mapreduce command was renamed to mapReduce in pymongo 3.9.0
                # mapreduce command consists of two mandatory parts: map and reduce
                cmd_doc = {
                    "map": event.command.get("map"),
                    "reduce": event.command.get("reduce")
                }

            if cmd_doc is not None:
                span.set_tag("json", json_util.dumps(cmd_doc))

    monitoring.register(MongoCommandTracer())

    logger.debug("Instrumenting pymongo")

except ImportError:
    pass



from __future__ import print_function
import os
import logging

logger = None


def get_standard_logger():
    """
    Retrieves and configures a standard logger for the Instana package

    @return: Logger
    """
    standard_logger = logging.getLogger("instana")

    ch = logging.StreamHandler()
    f = logging.Formatter('%(asctime)s: %(process)d %(levelname)s %(name)s: %(message)s')
    ch.setFormatter(f)
    standard_logger.addHandler(ch)
    standard_logger.setLevel(logging.DEBUG)
    return standard_logger


def get_aws_lambda_logger():
    """
    Retrieves the preferred logger for AWS Lambda

    @return: Logger
    """
    aws_lambda_logger = logging.getLogger()
    aws_lambda_logger.setLevel(logging.INFO)
    return aws_lambda_logger

def glogging_available():
    package_check = False

    # Is the glogging package available?
    try:
        from gunicorn import glogging
    except ImportError:
        pass
    else:
        package_check = True
    
    return package_check

def gunicorn_logging_available():
    from .util import running_in_gunicorn
    return running_in_gunicorn() and glogging_available()

aws_env = os.environ.get("AWS_EXECUTION_ENV", "")
env_is_aws_lambda = "AWS_Lambda_" in aws_env

if gunicorn_logging_available():
    logger = logging.getLogger("gunicorn.error")
elif env_is_aws_lambda is True:
    logger = get_aws_lambda_logger()
else:
    logger = get_standard_logger()



# Accept, process and queue spans for eventual reporting.
from __future__ import absolute_import

import os
import sys

from basictracer import Sampler

from .span import (RegisteredSpan, SDKSpan)

if sys.version_info.major == 2:
    import Queue as queue
else:
    import queue


class StanRecorder(object):
    THREAD_NAME = "Instana Span Reporting"

    REGISTERED_SPANS = ("aiohttp-client", "aiohttp-server", "asgi", "aws.lambda.entry", "boto3", "cassandra",
                        "celery-client", "celery-worker", "couchbase", "django", "gcs", "log",
                        "memcache", "mongo", "mysql", "postgres", "pymongo", "rabbitmq", "redis",
                        "render", "rpc-client", "rpc-server", "sqlalchemy", "soap", "tornado-client",
                        "tornado-server", "urllib3", "wsgi")

    # Recorder thread for collection/reporting of spans
    thread = None

    def __init__(self, agent = None):
        if agent is None:
            # Late import to avoid circular import
            # pylint: disable=import-outside-toplevel
            from .singletons import get_agent
            self.agent = get_agent()
        else:
            self.agent = agent

    def queue_size(self):
        """ Return the size of the queue; how may spans are queued, """
        return self.agent.collector.span_queue.qsize()

    def queued_spans(self):
        """ Get all of the spans in the queue """
        span = None
        spans = []

        import time
        from .singletons import env_is_test
        if env_is_test is True:
            time.sleep(1)

        if self.agent.collector.span_queue.empty() is True:
            return spans

        while True:
            try:
                span = self.agent.collector.span_queue.get(False)
            except queue.Empty:
                break
            else:
                spans.append(span)
        return spans

    def clear_spans(self):
        """ Clear the queue of spans """
        if self.agent.collector.span_queue.empty() == False:
            self.queued_spans()

    def record_span(self, span):
        """
        Convert the passed BasicSpan into and add it to the span queue
        """
        if self.agent.can_send():
            service_name = None
            source = self.agent.get_from_structure()
            if "INSTANA_SERVICE_NAME" in os.environ:
                service_name = self.agent.options.service_name

            if span.operation_name in self.REGISTERED_SPANS:
                json_span = RegisteredSpan(span, source, service_name)
            else:
                service_name = self.agent.options.service_name
                json_span = SDKSpan(span, source, service_name)

            # logger.debug("Recorded span: %s", json_span)
            self.agent.collector.span_queue.put(json_span)


class InstanaSampler(Sampler):
    def sampled(self, _):
        # We never sample
        return False



from __future__ import absolute_import

import sys

from ..log import logger
from .base_propagator import BasePropagator

PY2 = sys.version_info[0] == 2
PY3 = sys.version_info[0] == 3

class HTTPPropagator(BasePropagator):
    """
    Instana Propagator for Format.HTTP_HEADERS.

    The HTTP_HEADERS format deals with key-values with string to string mapping.
    The character set should be restricted to HTTP compatible.
    """
    def inject(self, span_context, carrier):
        try:
            trace_id = span_context.trace_id
            span_id = span_context.span_id

            if isinstance(carrier, dict) or hasattr(carrier, "__dict__"):
                carrier[self.HEADER_KEY_T] = trace_id
                carrier[self.HEADER_KEY_S] = span_id
                carrier[self.HEADER_KEY_L] = "1"
            elif isinstance(carrier, list):
                carrier.append((self.HEADER_KEY_T, trace_id))
                carrier.append((self.HEADER_KEY_S, span_id))
                carrier.append((self.HEADER_KEY_L, "1"))
            elif hasattr(carrier, '__setitem__'):
                carrier.__setitem__(self.HEADER_KEY_T, trace_id)
                carrier.__setitem__(self.HEADER_KEY_S, span_id)
                carrier.__setitem__(self.HEADER_KEY_L, "1")
            else:
                raise Exception("Unsupported carrier type", type(carrier))

        except Exception:
            logger.debug("inject error:", exc_info=True)



from __future__ import absolute_import

from ..log import logger
from .pep0249 import ConnectionFactory

try:
    import pymysql #

    cf = ConnectionFactory(connect_func=pymysql.connect, module_name='mysql')

    setattr(pymysql, 'connect', cf)
    if hasattr(pymysql, 'Connect'):
        setattr(pymysql, 'Connect', cf)

    logger.debug("Instrumenting pymysql")
except ImportError:
    pass



from fastapi import FastAPI, HTTPException
from fastapi.exceptions import RequestValidationError
from fastapi.responses import PlainTextResponse
from starlette.exceptions import HTTPException as StarletteHTTPException

fastapi_server = FastAPI()

# @fastapi_server.exception_handler(StarletteHTTPException)
# async def http_exception_handler(request, exc):
#     return PlainTextResponse(str(exc.detail), status_code=exc.status_code)

# @fastapi_server.exception_handler(RequestValidationError)
# async def validation_exception_handler(request, exc):
#     return PlainTextResponse(str(exc), status_code=400)

@fastapi_server.get("/")
async def root():
    import ipdb; ipdb.set_trace()
    return {"message": "Hello World"}

@fastapi_server.get("/users/{user_id}")
async def user(user_id):
    return {"user": user_id}

@fastapi_server.get("/400")
async def four_zero_zero():
    raise HTTPException(status_code=400, detail="400 response")

@fastapi_server.get("/404")
async def four_zero_four():
    raise HTTPException(status_code=404, detail="Item not found")

@fastapi_server.get("/500")
async def five_hundred():
    raise HTTPException(status_code=500, detail="500 response")

@fastapi_server.get("/starlette_exception")
async def starlette_exception():
    raise StarletteHTTPException(status_code=500, detail="500 response")


""" Module to assist in the data collection about the AWS Fargate task that is running this process """
from ....log import logger
from ..base import BaseHelper
from ....util import DictionaryOfStan


class TaskHelper(BaseHelper):
    """ This class helps in collecting data about the AWS Fargate task that is running """
    def collect_metrics(self, with_snapshot=False):
        """
        Collect and return metrics data (and optionally snapshot data) for this task
        @return: list - with one plugin entity
        """
        plugins = []

        try:
            if self.collector.task_metadata is not None:
                try:
                    plugin_data = dict()
                    plugin_data["name"] = "com.instana.plugin.aws.ecs.task"
                    plugin_data["entityId"] = self.collector.task_metadata.get("TaskARN", None)
                    plugin_data["data"] = DictionaryOfStan()
                    plugin_data["data"]["taskArn"] = self.collector.task_metadata.get("TaskARN", None)
                    plugin_data["data"]["clusterArn"] = self.collector.task_metadata.get("Cluster", None)
                    plugin_data["data"]["taskDefinition"] = self.collector.task_metadata.get("Family", None)
                    plugin_data["data"]["taskDefinitionVersion"] = self.collector.task_metadata.get("Revision", None)
                    plugin_data["data"]["availabilityZone"] = self.collector.task_metadata.get("AvailabilityZone", None)

                    if with_snapshot is True:
                        plugin_data["data"]["desiredStatus"] = self.collector.task_metadata.get("DesiredStatus", None)
                        plugin_data["data"]["knownStatus"] = self.collector.task_metadata.get("KnownStatus", None)
                        plugin_data["data"]["pullStartedAt"] = self.collector.task_metadata.get("PullStartedAt", None)
                        plugin_data["data"]["pullStoppedAt"] = self.collector.task_metadata.get("PullStoppeddAt", None)
                        limits = self.collector.task_metadata.get("Limits", {})
                        plugin_data["data"]["limits"]["cpu"] = limits.get("CPU", None)
                        plugin_data["data"]["limits"]["memory"] = limits.get("Memory", None)

                        if self.collector.agent.options.zone is not None:
                            plugin_data["data"]["instanaZone"] = self.collector.agent.options.zone

                        if self.collector.agent.options.tags is not None:
                            plugin_data["data"]["tags"] = self.collector.agent.options.tags
                except Exception:
                    logger.debug("collect_task_metrics: ", exc_info=True)
                finally:
                    plugins.append(plugin_data)
        except Exception:
            logger.debug("collect_task_metrics: ", exc_info=True)
        return plugins



"""
Instrumentation for FastAPI
https://fastapi.tiangolo.com/
"""
try:
    import fastapi
    import wrapt
    import signal
    import os

    from ..log import logger
    from ..util import running_in_gunicorn
    from .asgi import InstanaASGIMiddleware
    from starlette.middleware import Middleware

    @wrapt.patch_function_wrapper('fastapi.applications', 'FastAPI.__init__')
    def init_with_instana(wrapped, instance, args, kwargs):
        middleware = kwargs.get('middleware')
        if middleware is None:
            kwargs['middleware'] = [Middleware(InstanaASGIMiddleware)]
        elif isinstance(middleware, list):
            middleware.append(Middleware(InstanaASGIMiddleware))

        return wrapped(*args, **kwargs)

    logger.debug("Instrumenting FastAPI")

    # Reload GUnicorn 
    if "INSTANA_MAGIC" in os.environ and running_in_gunicorn():
        os.kill(os.getpid(), signal.SIGHUP)

except ImportError:
    pass


import uvicorn
from ...helpers import testenv
from instana.log import logger

testenv["fastapi_port"] = 10816
testenv["fastapi_server"] = ("http://127.0.0.1:" + str(testenv["fastapi_port"]))

def launch_fastapi():
    from .app import fastapi_server
    from instana.singletons import agent

    # Hack together a manual custom headers list; We'll use this in tests
    agent.options.extra_http_headers = [u'X-Capture-This', u'X-Capture-That']

    uvicorn.run(fastapi_server, host='127.0.0.1', port=testenv['fastapi_port'], log_level="critical")



import sys
import threading
import signal

from ...log import logger
from ..runtime import runtime_info
from ..profile import Profile
from ..profile import CallSite

if runtime_info.GEVENT:
    import gevent


class BlockSampler(object):
    SAMPLING_RATE = 0.05
    MAX_TRACEBACK_SIZE = 25 # number of frames

    def __init__(self, profiler):
        self.profiler = profiler
        self.ready = False
        self.top = None
        self.top_lock = threading.Lock()
        self.prev_signal_handler = None
        self.sampler_active = False

    def setup(self):
        if self.profiler.get_option('block_sampler_disabled'):
            return

        if not runtime_info.OS_LINUX and not runtime_info.OS_DARWIN:
            logger.debug('CPU profiler is only supported on Linux and OS X.')
            return

        sample_time = self.SAMPLING_RATE * 1000

        main_thread_id = None
        if runtime_info.GEVENT:
            main_thread_id = gevent._threading.get_ident()
        else:
            main_thread_id = threading.current_thread().ident

        def _sample(signum, signal_frame):
            if self.sampler_active:
                return
            self.sampler_active = True

            with self.top_lock:
                try:
                    self.process_sample(signal_frame, sample_time, main_thread_id)
                    signal_frame = None
                except Exception:
                    logger.error('Error processing sample', exc_info=True)

            self.sampler_active = False

        self.prev_signal_handler = signal.signal(signal.SIGALRM, _sample)

        self.ready = True

    def destroy(self):
        if not self.ready:
            return

        signal.signal(signal.SIGALRM, self.prev_signal_handler)

    def reset(self):
        self.top = CallSite('', '', 0)

    def start_sampler(self):
        logger.debug('Activating block sampler.')

        signal.setitimer(signal.ITIMER_REAL, self.SAMPLING_RATE, self.SAMPLING_RATE)

    def stop_sampler(self):
        signal.setitimer(signal.ITIMER_REAL, 0)

        logger.debug('Deactivating block sampler.')

    def build_profile(self, duration, timespan):
        with self.top_lock:
            self.top.normalize(duration)
            self.top.floor()

            profile = Profile(
                Profile.CATEGORY_TIME,
                Profile.TYPE_BLOCKING_CALLS,
                Profile.UNIT_MILLISECOND,
                self.top.children.values(),
                duration,
                timespan
            )

            return profile

    def process_sample(self, signal_frame, sample_time, main_thread_id):
        if self.top:
            current_frames = sys._current_frames()
            items = current_frames.items()
            for thread_id, thread_frame in items:
                if thread_id == main_thread_id:
                    thread_frame = signal_frame

                stack = self.recover_stack(thread_frame)
                if stack:
                    current_node = self.top
                    for func_name, filename, lineno in reversed(stack):
                        current_node = current_node.find_or_add_child(func_name, filename, lineno)
                    current_node.increment(sample_time, 1)

                thread_id, thread_frame, stack = None, None, None

            items = None
            current_frames = None


    def recover_stack(self, thread_frame):
        stack = []

        depth = 0
        while thread_frame is not None and depth <= self.MAX_TRACEBACK_SIZE:
            if thread_frame.f_code and thread_frame.f_code.co_name and thread_frame.f_code.co_filename:
                func_name = thread_frame.f_code.co_name
                filename = thread_frame.f_code.co_filename
                lineno = thread_frame.f_lineno

                if filename and self.profiler.frame_cache.is_profiler_frame(filename):
                    return None

                stack.append((func_name, filename, lineno))

                thread_frame = thread_frame.f_back

            depth += 1

        if len(stack) == 0:
            return None
        else:
            return stack



from __future__ import absolute_import

import sys
import unittest
import urllib3
from flask.signals import signals_available

import tests.apps.flask_app
from instana.singletons import tracer
from ..helpers import testenv


class TestFlask(unittest.TestCase):
    def setUp(self):
        """ Clear all spans before a test run """
        self.http = urllib3.PoolManager()
        self.recorder = tracer.recorder
        self.recorder.clear_spans()

    def tearDown(self):
        """ Do nothing for now """
        return None

    def test_vanilla_requests(self):
        r = self.http.request('GET', testenv["wsgi_server"] + '/')
        self.assertEqual(r.status, 200)

        spans = self.recorder.queued_spans()
        self.assertEqual(1, len(spans))

    def test_get_request(self):
        with tracer.start_active_span('test'):
            response = self.http.request('GET', testenv["wsgi_server"] + '/')

        spans = self.recorder.queued_spans()
        self.assertEqual(3, len(spans))

        wsgi_span = spans[0]
        urllib3_span = spans[1]
        test_span = spans[2]

        assert response
        self.assertEqual(200, response.status)

        assert('X-Instana-T' in response.headers)
        assert(int(response.headers['X-Instana-T'], 16))
        self.assertEqual(response.headers['X-Instana-T'], wsgi_span.t)

        assert('X-Instana-S' in response.headers)
        assert(int(response.headers['X-Instana-S'], 16))
        self.assertEqual(response.headers['X-Instana-S'], wsgi_span.s)

        assert('X-Instana-L' in response.headers)
        self.assertEqual(response.headers['X-Instana-L'], '1')

        assert('Server-Timing' in response.headers)
        server_timing_value = "intid;desc=%s" % wsgi_span.t
        self.assertEqual(response.headers['Server-Timing'], server_timing_value)

        self.assertIsNone(tracer.active_span)

        # Same traceId
        self.assertEqual(test_span.t, urllib3_span.t)
        self.assertEqual(urllib3_span.t, wsgi_span.t)

        # Parent relationships
        self.assertEqual(urllib3_span.p, test_span.s)
        self.assertEqual(wsgi_span.p, urllib3_span.s)

        # Synthetic
        self.assertIsNone(wsgi_span.sy)
        self.assertIsNone(urllib3_span.sy)
        self.assertIsNone(test_span.sy)

        # Error logging
        self.assertIsNone(test_span.ec)
        self.assertIsNone(urllib3_span.ec)
        self.assertIsNone(wsgi_span.ec)

        # wsgi
        self.assertEqual("wsgi", wsgi_span.n)
        self.assertEqual('127.0.0.1:' + str(testenv['wsgi_port']), wsgi_span.data["http"]["host"])
        self.assertEqual('/', wsgi_span.data["http"]["url"])
        self.assertEqual('GET', wsgi_span.data["http"]["method"])
        self.assertEqual(200, wsgi_span.data["http"]["status"])
        self.assertIsNone(wsgi_span.data["http"]["error"])
        self.assertIsNotNone(wsgi_span.stack)
        self.assertEqual(2, len(wsgi_span.stack))
        self.assertIsNone(wsgi_span.data['service'])

        # urllib3
        self.assertEqual("test", test_span.data["sdk"]["name"])
        self.assertEqual("urllib3", urllib3_span.n)
        self.assertEqual(200, urllib3_span.data["http"]["status"])
        self.assertEqual(testenv["wsgi_server"] + '/', urllib3_span.data["http"]["url"])
        self.assertEqual("GET", urllib3_span.data["http"]["method"])
        self.assertIsNotNone(urllib3_span.stack)
        self.assertTrue(type(urllib3_span.stack) is list)
        self.assertTrue(len(urllib3_span.stack) > 1)

        # We should NOT have a path template for this route
        self.assertIsNone(wsgi_span.data["http"]["path_tpl"])

    def test_synthetic_request(self):
        headers = {
            'X-Instana-Synthetic': '1'
        }

        with tracer.start_active_span('test'):
            response = self.http.request('GET', testenv["wsgi_server"] + '/', headers=headers)

        spans = self.recorder.queued_spans()
        self.assertEqual(3, len(spans))

        wsgi_span = spans[0]
        urllib3_span = spans[1]
        test_span = spans[2]

        self.assertTrue(wsgi_span.sy)
        self.assertIsNone(urllib3_span.sy)
        self.assertIsNone(test_span.sy)

    def test_render_template(self):
        with tracer.start_active_span('test'):
            response = self.http.request('GET', testenv["wsgi_server"] + '/render')

        spans = self.recorder.queued_spans()
        self.assertEqual(4, len(spans))

        render_span = spans[0]
        wsgi_span = spans[1]
        urllib3_span = spans[2]
        test_span = spans[3]

        assert response
        self.assertEqual(200, response.status)

        assert('X-Instana-T' in response.headers)
        assert(int(response.headers['X-Instana-T'], 16))
        self.assertEqual(response.headers['X-Instana-T'], wsgi_span.t)

        assert('X-Instana-S' in response.headers)
        assert(int(response.headers['X-Instana-S'], 16))
        self.assertEqual(response.headers['X-Instana-S'], wsgi_span.s)

        assert('X-Instana-L' in response.headers)
        self.assertEqual(response.headers['X-Instana-L'], '1')

        assert('Server-Timing' in response.headers)
        server_timing_value = "intid;desc=%s" % wsgi_span.t
        self.assertEqual(response.headers['Server-Timing'], server_timing_value)

        self.assertIsNone(tracer.active_span)

        # Same traceId
        self.assertEqual(test_span.t, render_span.t)
        self.assertEqual(test_span.t, urllib3_span.t)
        self.assertEqual(test_span.t, wsgi_span.t)

        # Parent relationships
        self.assertEqual(urllib3_span.p, test_span.s)
        self.assertEqual(wsgi_span.p, urllib3_span.s)
        self.assertEqual(render_span.p, wsgi_span.s)

        # Error logging
        self.assertIsNone(test_span.ec)
        self.assertIsNone(urllib3_span.ec)
        self.assertIsNone(wsgi_span.ec)
        self.assertIsNone(render_span.ec)

        # render
        self.assertEqual("render", render_span.n)
        self.assertEqual(3, render_span.k)
        self.assertEqual('flask_render_template.html', render_span.data["render"]["name"])
        self.assertEqual('template', render_span.data["render"]["type"])
        self.assertIsNone(render_span.data["log"]["message"])
        self.assertIsNone(render_span.data["log"]["parameters"])

        # wsgi
        self.assertEqual("wsgi", wsgi_span.n)
        self.assertEqual('127.0.0.1:' + str(testenv['wsgi_port']), wsgi_span.data["http"]["host"])
        self.assertEqual('/render', wsgi_span.data["http"]["url"])
        self.assertEqual('GET', wsgi_span.data["http"]["method"])
        self.assertEqual(200, wsgi_span.data["http"]["status"])
        self.assertIsNone(wsgi_span.data["http"]["error"])
        self.assertIsNotNone(wsgi_span.stack)
        self.assertEqual(2, len(wsgi_span.stack))
        self.assertIsNone(wsgi_span.data['service'])

        # urllib3
        self.assertEqual("test", test_span.data["sdk"]["name"])
        self.assertEqual("urllib3", urllib3_span.n)
        self.assertEqual(200, urllib3_span.data["http"]["status"])
        self.assertEqual(testenv["wsgi_server"] + '/render', urllib3_span.data["http"]["url"])
        self.assertEqual("GET", urllib3_span.data["http"]["method"])
        self.assertIsNotNone(urllib3_span.stack)
        self.assertTrue(type(urllib3_span.stack) is list)
        self.assertTrue(len(urllib3_span.stack) > 1)

        # We should NOT have a path template for this route
        self.assertIsNone(wsgi_span.data["http"]["path_tpl"])

    def test_render_template_string(self):
        with tracer.start_active_span('test'):
            response = self.http.request('GET', testenv["wsgi_server"] + '/render_string')

        spans = self.recorder.queued_spans()
        self.assertEqual(4, len(spans))

        render_span = spans[0]
        wsgi_span = spans[1]
        urllib3_span = spans[2]
        test_span = spans[3]

        assert response
        self.assertEqual(200, response.status)

        assert('X-Instana-T' in response.headers)
        assert(int(response.headers['X-Instana-T'], 16))
        self.assertEqual(response.headers['X-Instana-T'], wsgi_span.t)

        assert('X-Instana-S' in response.headers)
        assert(int(response.headers['X-Instana-S'], 16))
        self.assertEqual(response.headers['X-Instana-S'], wsgi_span.s)

        assert('X-Instana-L' in response.headers)
        self.assertEqual(response.headers['X-Instana-L'], '1')

        assert('Server-Timing' in response.headers)
        server_timing_value = "intid;desc=%s" % wsgi_span.t
        self.assertEqual(response.headers['Server-Timing'], server_timing_value)

        self.assertIsNone(tracer.active_span)

        # Same traceId
        self.assertEqual(test_span.t, render_span.t)
        self.assertEqual(test_span.t, urllib3_span.t)
        self.assertEqual(test_span.t, wsgi_span.t)

        # Parent relationships
        self.assertEqual(urllib3_span.p, test_span.s)
        self.assertEqual(wsgi_span.p, urllib3_span.s)
        self.assertEqual(render_span.p, wsgi_span.s)

        # Error logging
        self.assertIsNone(test_span.ec)
        self.assertIsNone(urllib3_span.ec)
        self.assertIsNone(wsgi_span.ec)
        self.assertIsNone(render_span.ec)

        # render
        self.assertEqual("render", render_span.n)
        self.assertEqual(3, render_span.k)
        self.assertEqual('(from string)', render_span.data["render"]["name"])
        self.assertEqual('template', render_span.data["render"]["type"])
        self.assertIsNone(render_span.data["log"]["message"])
        self.assertIsNone(render_span.data["log"]["parameters"])

        # wsgi
        self.assertEqual("wsgi", wsgi_span.n)
        self.assertEqual('127.0.0.1:' + str(testenv['wsgi_port']), wsgi_span.data["http"]["host"])
        self.assertEqual('/render_string', wsgi_span.data["http"]["url"])
        self.assertEqual('GET', wsgi_span.data["http"]["method"])
        self.assertEqual(200, wsgi_span.data["http"]["status"])
        self.assertIsNone(wsgi_span.data["http"]["error"])
        self.assertIsNotNone(wsgi_span.stack)
        self.assertEqual(2, len(wsgi_span.stack))
        self.assertIsNone(wsgi_span.data['service'])

        # urllib3
        self.assertEqual("test", test_span.data["sdk"]["name"])
        self.assertEqual("urllib3", urllib3_span.n)
        self.assertEqual(200, urllib3_span.data["http"]["status"])
        self.assertEqual(testenv["wsgi_server"] + '/render_string', urllib3_span.data["http"]["url"])
        self.assertEqual("GET", urllib3_span.data["http"]["method"])
        self.assertIsNotNone(urllib3_span.stack)
        self.assertTrue(type(urllib3_span.stack) is list)
        self.assertTrue(len(urllib3_span.stack) > 1)

        # We should NOT have a path template for this route
        self.assertIsNone(wsgi_span.data["http"]["path_tpl"])

    def test_301(self):
        with tracer.start_active_span('test'):
            response = self.http.request('GET', testenv["wsgi_server"] + '/301', redirect=False)

        spans = self.recorder.queued_spans()

        self.assertEqual(3, len(spans))

        wsgi_span = spans[0]
        urllib3_span = spans[1]
        test_span = spans[2]

        assert response
        self.assertEqual(301, response.status)

        assert('X-Instana-T' in response.headers)
        assert(int(response.headers['X-Instana-T'], 16))
        self.assertEqual(response.headers['X-Instana-T'], wsgi_span.t)

        assert('X-Instana-S' in response.headers)
        assert(int(response.headers['X-Instana-S'], 16))
        self.assertEqual(response.headers['X-Instana-S'], wsgi_span.s)

        assert('X-Instana-L' in response.headers)
        self.assertEqual(response.headers['X-Instana-L'], '1')

        assert('Server-Timing' in response.headers)
        server_timing_value = "intid;desc=%s" % wsgi_span.t
        self.assertEqual(response.headers['Server-Timing'], server_timing_value)

        self.assertIsNone(tracer.active_span)

        # Same traceId
        self.assertEqual(test_span.t, urllib3_span.t)
        self.assertEqual(test_span.t, wsgi_span.t)

        # Parent relationships
        self.assertEqual(urllib3_span.p, test_span.s)
        self.assertEqual(wsgi_span.p, urllib3_span.s)

        # Error logging
        self.assertIsNone(test_span.ec)
        self.assertEqual(None, urllib3_span.ec)
        self.assertEqual(None, wsgi_span.ec)

        # wsgi
        self.assertEqual("wsgi", wsgi_span.n)
        self.assertEqual('127.0.0.1:' + str(testenv['wsgi_port']), wsgi_span.data["http"]["host"])
        self.assertEqual('/301', wsgi_span.data["http"]["url"])
        self.assertEqual('GET', wsgi_span.data["http"]["method"])
        self.assertEqual(301, wsgi_span.data["http"]["status"])
        self.assertIsNone(wsgi_span.data["http"]["error"])
        self.assertIsNotNone(wsgi_span.stack)
        self.assertEqual(2, len(wsgi_span.stack))
        self.assertIsNone(wsgi_span.data['service'])

        # urllib3
        self.assertEqual("test", test_span.data["sdk"]["name"])
        self.assertEqual("urllib3", urllib3_span.n)
        self.assertEqual(301, urllib3_span.data["http"]["status"])
        self.assertEqual(testenv["wsgi_server"] + '/301', urllib3_span.data["http"]["url"])
        self.assertEqual("GET", urllib3_span.data["http"]["method"])
        self.assertIsNotNone(urllib3_span.stack)
        self.assertTrue(type(urllib3_span.stack) is list)
        self.assertTrue(len(urllib3_span.stack) > 1)

        # We should NOT have a path template for this route
        self.assertIsNone(wsgi_span.data["http"]["path_tpl"])

    def test_custom_404(self):
        with tracer.start_active_span('test'):
            response = self.http.request('GET', testenv["wsgi_server"] + '/custom-404')

        spans = self.recorder.queued_spans()

        self.assertEqual(3, len(spans))

        wsgi_span = spans[0]
        urllib3_span = spans[1]
        test_span = spans[2]

        assert response
        self.assertEqual(404, response.status)

        # assert('X-Instana-T' in response.headers)
        # assert(int(response.headers['X-Instana-T'], 16))
        # self.assertEqual(response.headers['X-Instana-T'], wsgi_span.t)
        #
        # assert('X-Instana-S' in response.headers)
        # assert(int(response.headers['X-Instana-S'], 16))
        # self.assertEqual(response.headers['X-Instana-S'], wsgi_span.s)
        #
        # assert('X-Instana-L' in response.headers)
        # self.assertEqual(response.headers['X-Instana-L'], '1')
        #
        # assert('Server-Timing' in response.headers)
        # server_timing_value = "intid;desc=%s" % wsgi_span.t
        # self.assertEqual(response.headers['Server-Timing'], server_timing_value)

        self.assertIsNone(tracer.active_span)

        # Same traceId
        self.assertEqual(test_span.t, urllib3_span.t)
        self.assertEqual(test_span.t, wsgi_span.t)

        # Parent relationships
        self.assertEqual(urllib3_span.p, test_span.s)
        self.assertEqual(wsgi_span.p, urllib3_span.s)

        # Error logging
        self.assertIsNone(test_span.ec)
        self.assertEqual(None, urllib3_span.ec)
        self.assertEqual(None, wsgi_span.ec)

        # wsgi
        self.assertEqual("wsgi", wsgi_span.n)
        self.assertEqual('127.0.0.1:' + str(testenv['wsgi_port']), wsgi_span.data["http"]["host"])
        self.assertEqual('/custom-404', wsgi_span.data["http"]["url"])
        self.assertEqual('GET', wsgi_span.data["http"]["method"])
        self.assertEqual(404, wsgi_span.data["http"]["status"])
        self.assertIsNone(wsgi_span.data["http"]["error"])
        self.assertIsNotNone(wsgi_span.stack)
        self.assertEqual(2, len(wsgi_span.stack))
        self.assertIsNone(wsgi_span.data['service'])

        # urllib3
        self.assertEqual("test", test_span.data["sdk"]["name"])
        self.assertEqual("urllib3", urllib3_span.n)
        self.assertEqual(404, urllib3_span.data["http"]["status"])
        self.assertEqual(testenv["wsgi_server"] + '/custom-404', urllib3_span.data["http"]["url"])
        self.assertEqual("GET", urllib3_span.data["http"]["method"])
        self.assertIsNotNone(urllib3_span.stack)
        self.assertTrue(type(urllib3_span.stack) is list)
        self.assertTrue(len(urllib3_span.stack) > 1)

        # We should NOT have a path template for this route
        self.assertIsNone(wsgi_span.data["http"]["path_tpl"])

    def test_404(self):
        with tracer.start_active_span('test'):
            response = self.http.request('GET', testenv["wsgi_server"] + '/11111111111')

        spans = self.recorder.queued_spans()

        self.assertEqual(3, len(spans))

        wsgi_span = spans[0]
        urllib3_span = spans[1]
        test_span = spans[2]

        assert response
        self.assertEqual(404, response.status)

        # assert('X-Instana-T' in response.headers)
        # assert(int(response.headers['X-Instana-T'], 16))
        # self.assertEqual(response.headers['X-Instana-T'], wsgi_span.t)
        #
        # assert('X-Instana-S' in response.headers)
        # assert(int(response.headers['X-Instana-S'], 16))
        # self.assertEqual(response.headers['X-Instana-S'], wsgi_span.s)
        #
        # assert('X-Instana-L' in response.headers)
        # self.assertEqual(response.headers['X-Instana-L'], '1')
        #
        # assert('Server-Timing' in response.headers)
        # server_timing_value = "intid;desc=%s" % wsgi_span.t
        # self.assertEqual(response.headers['Server-Timing'], server_timing_value)

        self.assertIsNone(tracer.active_span)

        # Same traceId
        self.assertEqual(test_span.t, urllib3_span.t)
        self.assertEqual(test_span.t, wsgi_span.t)

        # Parent relationships
        self.assertEqual(urllib3_span.p, test_span.s)
        self.assertEqual(wsgi_span.p, urllib3_span.s)

        # Error logging
        self.assertIsNone(test_span.ec)
        self.assertEqual(None, urllib3_span.ec)
        self.assertEqual(None, wsgi_span.ec)

        # wsgi
        self.assertEqual("wsgi", wsgi_span.n)
        self.assertEqual('127.0.0.1:' + str(testenv['wsgi_port']), wsgi_span.data["http"]["host"])
        self.assertEqual('/11111111111', wsgi_span.data["http"]["url"])
        self.assertEqual('GET', wsgi_span.data["http"]["method"])
        self.assertEqual(404, wsgi_span.data["http"]["status"])
        self.assertIsNone(wsgi_span.data["http"]["error"])
        self.assertIsNotNone(wsgi_span.stack)
        self.assertEqual(2, len(wsgi_span.stack))
        self.assertIsNone(wsgi_span.data['service'])

        # urllib3
        self.assertEqual("test", test_span.data["sdk"]["name"])
        self.assertEqual("urllib3", urllib3_span.n)
        self.assertEqual(404, urllib3_span.data["http"]["status"])
        self.assertEqual(testenv["wsgi_server"] + '/11111111111', urllib3_span.data["http"]["url"])
        self.assertEqual("GET", urllib3_span.data["http"]["method"])
        self.assertIsNotNone(urllib3_span.stack)
        self.assertTrue(type(urllib3_span.stack) is list)
        self.assertTrue(len(urllib3_span.stack) > 1)

        # We should NOT have a path template for this route
        self.assertIsNone(wsgi_span.data["http"]["path_tpl"])

    def test_500(self):
        with tracer.start_active_span('test'):
            response = self.http.request('GET', testenv["wsgi_server"] + '/500')

        spans = self.recorder.queued_spans()

        self.assertEqual(3, len(spans))

        wsgi_span = spans[0]
        urllib3_span = spans[1]
        test_span = spans[2]

        assert response
        self.assertEqual(500, response.status)

        assert('X-Instana-T' in response.headers)
        assert(int(response.headers['X-Instana-T'], 16))
        self.assertEqual(response.headers['X-Instana-T'], wsgi_span.t)

        assert('X-Instana-S' in response.headers)
        assert(int(response.headers['X-Instana-S'], 16))
        self.assertEqual(response.headers['X-Instana-S'], wsgi_span.s)

        assert('X-Instana-L' in response.headers)
        self.assertEqual(response.headers['X-Instana-L'], '1')

        assert('Server-Timing' in response.headers)
        server_timing_value = "intid;desc=%s" % wsgi_span.t
        self.assertEqual(response.headers['Server-Timing'], server_timing_value)

        self.assertIsNone(tracer.active_span)

        # Same traceId
        self.assertEqual(test_span.t, urllib3_span.t)
        self.assertEqual(test_span.t, wsgi_span.t)

        # Parent relationships
        self.assertEqual(urllib3_span.p, test_span.s)
        self.assertEqual(wsgi_span.p, urllib3_span.s)

        # Error logging
        self.assertIsNone(test_span.ec)
        self.assertEqual(1, urllib3_span.ec)
        self.assertEqual(1, wsgi_span.ec)

        # wsgi
        self.assertEqual("wsgi", wsgi_span.n)
        self.assertEqual('127.0.0.1:' + str(testenv['wsgi_port']), wsgi_span.data["http"]["host"])
        self.assertEqual('/500', wsgi_span.data["http"]["url"])
        self.assertEqual('GET', wsgi_span.data["http"]["method"])
        self.assertEqual(500, wsgi_span.data["http"]["status"])
        self.assertIsNone(wsgi_span.data["http"]["error"])
        self.assertIsNotNone(wsgi_span.stack)
        self.assertEqual(2, len(wsgi_span.stack))
        self.assertIsNone(wsgi_span.data['service'])

        # urllib3
        self.assertEqual("test", test_span.data["sdk"]["name"])
        self.assertEqual("urllib3", urllib3_span.n)
        self.assertEqual(500, urllib3_span.data["http"]["status"])
        self.assertEqual(testenv["wsgi_server"] + '/500', urllib3_span.data["http"]["url"])
        self.assertEqual("GET", urllib3_span.data["http"]["method"])
        self.assertIsNotNone(urllib3_span.stack)
        self.assertTrue(type(urllib3_span.stack) is list)
        self.assertTrue(len(urllib3_span.stack) > 1)

        # We should NOT have a path template for this route
        self.assertIsNone(wsgi_span.data["http"]["path_tpl"])

    def test_render_error(self):
        if signals_available is True:
            raise unittest.SkipTest("Exceptions without handlers vary with blinker")

        with tracer.start_active_span('test'):
            response = self.http.request('GET', testenv["wsgi_server"] + '/render_error')

        spans = self.recorder.queued_spans()

        self.assertEqual(4, len(spans))

        log_span = spans[0]
        wsgi_span = spans[1]
        urllib3_span = spans[2]
        test_span = spans[3]

        assert response
        self.assertEqual(500, response.status)

        # assert('X-Instana-T' in response.headers)
        # assert(int(response.headers['X-Instana-T'], 16))
        # self.assertEqual(response.headers['X-Instana-T'], wsgi_span.t)
        #
        # assert('X-Instana-S' in response.headers)
        # assert(int(response.headers['X-Instana-S'], 16))
        # self.assertEqual(response.headers['X-Instana-S'], wsgi_span.s)
        #
        # assert('X-Instana-L' in response.headers)
        # self.assertEqual(response.headers['X-Instana-L'], '1')
        #
        # assert('Server-Timing' in response.headers)
        # server_timing_value = "intid;desc=%s" % wsgi_span.t
        # self.assertEqual(response.headers['Server-Timing'], server_timing_value)

        self.assertIsNone(tracer.active_span)

        # Same traceId
        self.assertEqual(test_span.t, urllib3_span.t)
        self.assertEqual(test_span.t, wsgi_span.t)

        # Parent relationships
        self.assertEqual(urllib3_span.p, test_span.s)
        self.assertEqual(wsgi_span.p, urllib3_span.s)

        # Error logging
        self.assertIsNone(test_span.ec)
        self.assertEqual(1, urllib3_span.ec)
        self.assertEqual(1, wsgi_span.ec)

        # error log
        self.assertEqual("log", log_span.n)
        self.assertEqual('Exception on /render_error [GET]', log_span.data["log"]['message'])
        self.assertEqual("<class 'jinja2.exceptions.TemplateSyntaxError'> unexpected '}'", log_span.data["log"]['parameters'])

        # wsgi
        self.assertEqual("wsgi", wsgi_span.n)
        self.assertEqual('127.0.0.1:' + str(testenv['wsgi_port']), wsgi_span.data["http"]["host"])
        self.assertEqual('/render_error', wsgi_span.data["http"]["url"])
        self.assertEqual('GET', wsgi_span.data["http"]["method"])
        self.assertEqual(500, wsgi_span.data["http"]["status"])
        self.assertIsNone(wsgi_span.data["http"]["error"])
        self.assertIsNotNone(wsgi_span.stack)
        self.assertEqual(2, len(wsgi_span.stack))
        self.assertIsNone(wsgi_span.data['service'])

        # urllib3
        self.assertEqual("test", test_span.data["sdk"]["name"])
        self.assertEqual("urllib3", urllib3_span.n)
        self.assertEqual(500, urllib3_span.data["http"]["status"])
        self.assertEqual(testenv["wsgi_server"] + '/render_error', urllib3_span.data["http"]["url"])
        self.assertEqual("GET", urllib3_span.data["http"]["method"])
        self.assertIsNotNone(urllib3_span.stack)
        self.assertTrue(type(urllib3_span.stack) is list)
        self.assertTrue(len(urllib3_span.stack) > 1)

        # We should NOT have a path template for this route
        self.assertIsNone(wsgi_span.data["http"]["path_tpl"])

    def test_exception(self):
        if signals_available is True:
            raise unittest.SkipTest("Exceptions without handlers vary with blinker")

        with tracer.start_active_span('test'):
            response = self.http.request('GET', testenv["wsgi_server"] + '/exception')

        spans = self.recorder.queued_spans()

        self.assertEqual(4, len(spans))

        log_span = spans[0]
        wsgi_span = spans[1]
        urllib3_span = spans[2]
        test_span = spans[3]

        assert response
        self.assertEqual(500, response.status)

        self.assertIsNone(tracer.active_span)

        # Same traceId
        self.assertEqual(test_span.t, urllib3_span.t)
        self.assertEqual(test_span.t, wsgi_span.t)

        # Parent relationships
        self.assertEqual(urllib3_span.p, test_span.s)
        self.assertEqual(wsgi_span.p, urllib3_span.s)
        self.assertEqual(log_span.p, wsgi_span.s)

        # Error logging
        self.assertIsNone(test_span.ec)
        self.assertEqual(1, urllib3_span.ec)
        self.assertEqual(1, wsgi_span.ec)
        self.assertEqual(1, log_span.ec)

        # error log
        self.assertEqual("log", log_span.n)
        self.assertEqual('Exception on /exception [GET]', log_span.data["log"]['message'])
        if sys.version_info < (3, 0):
            self.assertEqual("<type 'exceptions.Exception'> fake error", log_span.data["log"]['parameters'])
        else:
            self.assertEqual("<class 'Exception'> fake error", log_span.data["log"]['parameters'])


        # wsgis
        self.assertEqual("wsgi", wsgi_span.n)
        self.assertEqual('127.0.0.1:' + str(testenv['wsgi_port']), wsgi_span.data["http"]["host"])
        self.assertEqual('/exception', wsgi_span.data["http"]["url"])
        self.assertEqual('GET', wsgi_span.data["http"]["method"])
        self.assertEqual(500, wsgi_span.data["http"]["status"])
        self.assertIsNone(wsgi_span.data["http"]["error"])
        self.assertIsNotNone(wsgi_span.stack)
        self.assertEqual(2, len(wsgi_span.stack))
        self.assertIsNone(wsgi_span.data['service'])

        # urllib3
        self.assertEqual("test", test_span.data["sdk"]["name"])
        self.assertEqual("urllib3", urllib3_span.n)
        self.assertEqual(500, urllib3_span.data["http"]["status"])
        self.assertEqual(testenv["wsgi_server"] + '/exception', urllib3_span.data["http"]["url"])
        self.assertEqual("GET", urllib3_span.data["http"]["method"])
        self.assertIsNotNone(urllib3_span.stack)
        self.assertTrue(type(urllib3_span.stack) is list)
        self.assertTrue(len(urllib3_span.stack) > 1)

        # We should NOT have a path template for this route
        self.assertIsNone(wsgi_span.data["http"]["path_tpl"])

    def test_custom_exception_with_log(self):
        with tracer.start_active_span('test'):
            response = self.http.request('GET', testenv["wsgi_server"] + '/exception-invalid-usage')

        spans = self.recorder.queued_spans()

        self.assertEqual(4, len(spans))

        log_span = spans[0]
        wsgi_span = spans[1]
        urllib3_span = spans[2]
        test_span = spans[3]

        assert response
        self.assertEqual(502, response.status)

        assert('X-Instana-T' in response.headers)
        assert(int(response.headers['X-Instana-T'], 16))
        self.assertEqual(response.headers['X-Instana-T'], wsgi_span.t)

        assert('X-Instana-S' in response.headers)
        assert(int(response.headers['X-Instana-S'], 16))
        self.assertEqual(response.headers['X-Instana-S'], wsgi_span.s)

        assert('X-Instana-L' in response.headers)
        self.assertEqual(response.headers['X-Instana-L'], '1')

        assert('Server-Timing' in response.headers)
        server_timing_value = "intid;desc=%s" % wsgi_span.t
        self.assertEqual(response.headers['Server-Timing'], server_timing_value)

        self.assertIsNone(tracer.active_span)

        # Same traceId
        self.assertEqual(test_span.t, urllib3_span.t)
        self.assertEqual(test_span.t, wsgi_span.t)

        # Parent relationships
        self.assertEqual(urllib3_span.p, test_span.s)
        self.assertEqual(wsgi_span.p, urllib3_span.s)

        # Error logging
        self.assertIsNone(test_span.ec)
        self.assertEqual(1, urllib3_span.ec)
        self.assertEqual(1, wsgi_span.ec)
        self.assertEqual(1, log_span.ec)

        # error log
        self.assertEqual("log", log_span.n)
        self.assertEqual('InvalidUsage error handler invoked', log_span.data["log"]['message'])
        self.assertEqual("<class 'tests.apps.flask_app.app.InvalidUsage'> ", log_span.data["log"]['parameters'])

        # wsgi
        self.assertEqual("wsgi", wsgi_span.n)
        self.assertEqual('127.0.0.1:' + str(testenv['wsgi_port']), wsgi_span.data["http"]["host"])
        self.assertEqual('/exception-invalid-usage', wsgi_span.data["http"]["url"])
        self.assertEqual('GET', wsgi_span.data["http"]["method"])
        self.assertEqual(502, wsgi_span.data["http"]["status"])
        self.assertEqual('Simulated custom exception', wsgi_span.data["http"]["error"])
        self.assertIsNotNone(wsgi_span.stack)
        self.assertEqual(2, len(wsgi_span.stack))
        self.assertIsNone(wsgi_span.data['service'])

        # urllib3
        self.assertEqual("test", test_span.data["sdk"]["name"])
        self.assertEqual("urllib3", urllib3_span.n)
        self.assertEqual(502, urllib3_span.data["http"]["status"])
        self.assertEqual(testenv["wsgi_server"] + '/exception-invalid-usage', urllib3_span.data["http"]["url"])
        self.assertEqual("GET", urllib3_span.data["http"]["method"])
        self.assertIsNotNone(urllib3_span.stack)
        self.assertTrue(type(urllib3_span.stack) is list)
        self.assertTrue(len(urllib3_span.stack) > 1)

        # We should NOT have a path template for this route
        self.assertIsNone(wsgi_span.data["http"]["path_tpl"])

    def test_path_templates(self):
        with tracer.start_active_span('test'):
            response = self.http.request('GET', testenv["wsgi_server"] + '/users/Ricky/sayhello')

        spans = self.recorder.queued_spans()
        self.assertEqual(3, len(spans))

        wsgi_span = spans[0]
        urllib3_span = spans[1]
        test_span = spans[2]

        assert response
        self.assertEqual(200, response.status)

        assert('X-Instana-T' in response.headers)
        assert(int(response.headers['X-Instana-T'], 16))
        self.assertEqual(response.headers['X-Instana-T'], wsgi_span.t)

        assert('X-Instana-S' in response.headers)
        assert(int(response.headers['X-Instana-S'], 16))
        self.assertEqual(response.headers['X-Instana-S'], wsgi_span.s)

        assert('X-Instana-L' in response.headers)
        self.assertEqual(response.headers['X-Instana-L'], '1')

        assert('Server-Timing' in response.headers)
        server_timing_value = "intid;desc=%s" % wsgi_span.t
        self.assertEqual(response.headers['Server-Timing'], server_timing_value)

        self.assertIsNone(tracer.active_span)

        # Same traceId
        self.assertEqual(test_span.t, urllib3_span.t)
        self.assertEqual(urllib3_span.t, wsgi_span.t)

        # Parent relationships
        self.assertEqual(urllib3_span.p, test_span.s)
        self.assertEqual(wsgi_span.p, urllib3_span.s)

        # Error logging
        self.assertIsNone(test_span.ec)
        self.assertIsNone(urllib3_span.ec)
        self.assertIsNone(wsgi_span.ec)

        # wsgi
        self.assertEqual("wsgi", wsgi_span.n)
        self.assertEqual('127.0.0.1:' + str(testenv['wsgi_port']), wsgi_span.data["http"]["host"])
        self.assertEqual('/users/Ricky/sayhello', wsgi_span.data["http"]["url"])
        self.assertEqual('GET', wsgi_span.data["http"]["method"])
        self.assertEqual(200, wsgi_span.data["http"]["status"])
        self.assertIsNone(wsgi_span.data["http"]["error"])
        self.assertIsNotNone(wsgi_span.stack)
        self.assertEqual(2, len(wsgi_span.stack))
        self.assertIsNone(wsgi_span.data['service'])

        # urllib3
        self.assertEqual("test", test_span.data["sdk"]["name"])
        self.assertEqual("urllib3", urllib3_span.n)
        self.assertEqual(200, urllib3_span.data["http"]["status"])
        self.assertEqual(testenv["wsgi_server"] + '/users/Ricky/sayhello', urllib3_span.data["http"]["url"])
        self.assertEqual("GET", urllib3_span.data["http"]["method"])
        self.assertIsNotNone(urllib3_span.stack)
        self.assertTrue(type(urllib3_span.stack) is list)
        self.assertTrue(len(urllib3_span.stack) > 1)

        # We should have a reported path template for this route
        self.assertEqual("/users/{username}/sayhello", wsgi_span.data["http"]["path_tpl"])




import threading
import signal

from ...log import logger
from ..runtime import runtime_info
from ..profile import Profile
from ..profile import CallSite


class CPUSampler(object):
    SAMPLING_RATE = 0.01
    MAX_TRACEBACK_SIZE = 25 # number of frames

    def __init__(self, profiler):
        self.profiler = profiler
        self.ready = False
        self.top = None
        self.top_lock = threading.Lock()
        self.prev_signal_handler = None
        self.sampler_active = False

    def setup(self):
        if self.profiler.get_option('cpu_sampler_disabled'):
            return

        if not runtime_info.OS_LINUX and not runtime_info.OS_DARWIN:
            logger.debug('CPU sampler is only supported on Linux and OS X.')
            return

        def _sample(signum, signal_frame):
            if self.sampler_active:
                return
            self.sampler_active = True

            with self.top_lock:
                try:
                    self.process_sample(signal_frame)
                    signal_frame = None
                except Exception:
                    logger.error('Error in signal handler', exc_info=True)
 
            self.sampler_active = False

        self.prev_signal_handler = signal.signal(signal.SIGPROF, _sample)

        self.ready = True

    def reset(self):
        self.top = CallSite('', '', 0)

    def start_sampler(self):
        logger.debug('Activating CPU sampler.')

        signal.setitimer(signal.ITIMER_PROF, self.SAMPLING_RATE, self.SAMPLING_RATE)

    def stop_sampler(self):
        signal.setitimer(signal.ITIMER_PROF, 0)

    def destroy(self):
        if not self.ready:
            return

        signal.signal(signal.SIGPROF, self.prev_signal_handler)

    def build_profile(self, duration, timespan):
        with self.top_lock:
            profile = Profile(
                Profile.CATEGORY_CPU,
                Profile.TYPE_CPU_USAGE,
                Profile.UNIT_SAMPLE,
                self.top.children.values(),
                duration,
                timespan
            )

            return profile

    def process_sample(self, signal_frame):
        if self.top:
            if signal_frame:
                stack = self.recover_stack(signal_frame)
                if stack:
                    self.update_profile(self.top, stack)

                stack = None

    def recover_stack(self, signal_frame):
        stack = []

        depth = 0
        while signal_frame is not None and depth <= self.MAX_TRACEBACK_SIZE:
            if signal_frame.f_code and signal_frame.f_code.co_name and signal_frame.f_code.co_filename:
                func_name = signal_frame.f_code.co_name
                filename = signal_frame.f_code.co_filename
                lineno = signal_frame.f_lineno

                if filename and self.profiler.frame_cache.is_profiler_frame(filename):
                    return None

                #frame = Frame(func_name, filename, lineno)
                stack.append((func_name, filename, lineno))

                signal_frame = signal_frame.f_back
            
            depth += 1

        if len(stack) == 0:
            return None
        else:
            return stack

    def update_profile(self, profile, stack):
        current_node = profile

        for func_name, filename, lineno in reversed(stack):
            current_node = current_node.find_or_add_child(func_name, filename, lineno)
        
        current_node.increment(1, 1)



# coding: utf-8
import os
import sys
from os import path
from distutils.version import LooseVersion
from setuptools import find_packages, setup

os.environ["INSTANA_DISABLE"] = "true"

# pylint: disable=wrong-import-position
from instana.version import VERSION

# Import README.md into long_description
pwd = path.abspath(path.dirname(__file__))

if sys.version_info[0] > 2:
    with open(path.join(pwd, 'README.md'), encoding='utf-8') as f:
        long_description = f.read()
else:
    with open(path.join(pwd, 'README.md')) as f:
        long_description = f.read()


def check_setuptools():
    """ Validate that we have min version required of setuptools """
    import pkg_resources
    st_version = pkg_resources.get_distribution('setuptools').version
    if LooseVersion(st_version) < LooseVersion('20.2.2'):
        exit('The Instana sensor requires a newer verion of `setuptools` (>=20.2.2).\n'
             'Please run `pip install --upgrade setuptools` to upgrade. \n'
             '  and then try the install again.\n'
             'Also:\n'
             '  `pip show setuptools` - shows the current version\n'
             '  To see the setuptools releases: \n'
             '    https://setuptools.readthedocs.io/en/latest/history.html')


check_setuptools()

setup(name='instana',
      version=VERSION,
      url='https://www.instana.com/',
      project_urls={
          'CI: CircleCI': 'https://circleci.com/gh/instana/python-sensor',
          'Documentation': 'https://docs.instana.io/ecosystem/python/',
          'GitHub: issues': 'https://github.com/instana/python-sensor/issues',
          'GitHub: repo': 'https://github.com/instana/python-sensor',
          'Support': 'https://support.instana.com',
      },
      license='MIT',
      author='Instana Inc.',
      author_email='peter.lombardo@instana.com',
      description='🐍 Python Distributed Tracing & Metrics Sensor for Instana',
      options={"bdist_wheel": {"universal": True}},
      packages=find_packages(exclude=['tests', 'examples']),
      long_description=long_description,
      long_description_content_type='text/markdown',
      zip_safe=False,
      install_requires=['autowrapt>=1.0',
                        'basictracer>=3.1.0',
                        'certifi>=2018.4.16',
                        'fysom>=2.1.2',
                        'opentracing>=2.3.0',
                        'requests>=2.8.0',
                        'six>=1.12.0',
                        'urllib3>=1.18.1'],
      entry_points={
                    'instana':  ['string = instana:load'],
                    'flask':    ['string = instana:load'],  # deprecated: use same as 'instana'
                    'runtime':  ['string = instana:load'],  # deprecated: use same as 'instana'
                    'django':   ['string = instana:load'],  # deprecated: use same as 'instana'
                    'django19': ['string = instana:load'],  # deprecated: use same as 'instana'
                    },
      extras_require={
          'test-gevent': [
              'flask>=0.12.2',
              'gevent>=1.4.0',
              'mock>=2.0.0',
              'nose>=1.0',
              'pyramid>=1.2',
              'pytest>=4.6',
              'urllib3[secure]>=1.15'
          ],
          'test-cassandra': [
              'cassandra-driver==3.20.2',
              'mock>=2.0.0',
              'nose>=1.0',
              'pytest>=4.6',
              'urllib3[secure]>=1.15'
          ],
          'test-couchbase': [
              'couchbase==2.5.9',
          ],
          'test': [
              'aiofiles>=0.5.0;python_version>="3.5"',
              'aiohttp>=3.5.4;python_version>="3.5"',
              'asynqp>=0.4;python_version>="3.5"',
              'boto3>=1.10.0',
              'celery>=4.1.1',
              'django>=1.11,<2.2',
              'fastapi>=0.61.1;python_version>="3.6"',
              'flask>=0.12.2',
              'grpcio>=1.18.0',
              'google-cloud-storage>=1.24.0;python_version>="3.5"',
              'lxml>=3.4',
              'mock>=2.0.0',
              'moto>=1.3.16',
              'mysqlclient>=1.3.14;python_version>="3.5"',
              'MySQL-python>=1.2.5;python_version<="2.7"',
              'nose>=1.0',
              'PyMySQL[rsa]>=0.9.1',
              'pyOpenSSL>=16.1.0;python_version<="2.7"',
              'psycopg2>=2.7.1',
              'pymongo>=3.7.0',
              'pyramid>=1.2',
              'pytest>=4.6',
              'pytest-celery',
              'redis>3.0.0',
              'requests>=2.17.1',
              'sqlalchemy>=1.1.15',
              'spyne>=2.9,<=2.12.14',
              'suds-jurko>=0.6',
              'tornado>=4.5.3,<6.0',
              'uvicorn>=0.12.2;python_version>="3.6"',
              'urllib3[secure]>=1.15'
          ],
      },
      test_suite='nose.collector',
      keywords=['performance', 'opentracing', 'metrics', 'monitoring',
                'tracing', 'distributed-tracing'],
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Framework :: Django',
          'Framework :: Flask',
          'Framework :: Pyramid',
          'Intended Audience :: Developers',
          'Intended Audience :: Information Technology',
          'Intended Audience :: Science/Research',
          'Intended Audience :: System Administrators',
          'License :: OSI Approved :: MIT License',
          'Operating System :: OS Independent',
          'Programming Language :: Python',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Topic :: Internet :: WWW/HTTP :: WSGI :: Middleware',
          'Topic :: System :: Monitoring',
          'Topic :: System :: Networking :: Monitoring',
          'Topic :: Software Development :: Libraries :: Python Modules'])



import threading
import time

from ..log import logger


class TimerWraper(object):
    def __init__(self):
        self.timer = None
        self.cancel_lock = threading.Lock()
        self.canceled = False

    def cancel(self):
        with self.cancel_lock:
            self.canceled = True
            self.timer.cancel()


def delay(timeout, func, *args):
    def func_wrapper():
        try:
            func(*args)
        except Exception:
            logger.error('Error in delayed function', exc_info=True)

    t = threading.Timer(timeout, func_wrapper, ())
    t.start()

    return t


def schedule( timeout, interval, func, *args):
    tw = TimerWraper()

    def func_wrapper():
        start = time.time()

        try:
            func(*args)
        except Exception:
            logger.error('Error in scheduled function', exc_info=True)

        with tw.cancel_lock:
            if not tw.canceled:
                tw.timer = threading.Timer(abs(interval - (time.time() - start)), func_wrapper, ())
                tw.timer.start()

    tw.timer = threading.Timer(timeout, func_wrapper, ())
    tw.timer.start()

    return tw


"""
AWS Fargate Collector: Manages the periodic collection of metrics & snapshot data
"""
import os
import json
from time import time
import requests

from ..log import logger
from .base import BaseCollector
from ..util import DictionaryOfStan, validate_url
from ..singletons import env_is_test

from .helpers.process import ProcessHelper
from .helpers.runtime import RuntimeHelper
from .helpers.fargate.task import TaskHelper
from .helpers.fargate.docker import DockerHelper
from .helpers.fargate.container import ContainerHelper


class AWSFargateCollector(BaseCollector):
    """ Collector for AWS Fargate """
    def __init__(self, agent):
        super(AWSFargateCollector, self).__init__(agent)
        logger.debug("Loading AWS Fargate Collector")

        # Indicates if this Collector has all requirements to run successfully
        self.ready_to_start = True

        # Prepare the URLS that we will collect data from
        self.ecmu = os.environ.get("ECS_CONTAINER_METADATA_URI", "")

        if self.ecmu == "" or validate_url(self.ecmu) is False:
            logger.warning("AWSFargateCollector: ECS_CONTAINER_METADATA_URI not in environment or invalid URL.  "
                           "Instana will not be able to monitor this environment")
            self.ready_to_start = False

        self.ecmu_url_root = self.ecmu
        self.ecmu_url_task = self.ecmu + '/task'
        self.ecmu_url_stats = self.ecmu + '/stats'
        self.ecmu_url_task_stats = self.ecmu + '/task/stats'

        # Timestamp in seconds of the last time we fetched all ECMU data
        self.last_ecmu_full_fetch = 0

        # How often to do a full fetch of ECMU data
        self.ecmu_full_fetch_interval = 304

        # HTTP client with keep-alive
        self.http_client = requests.Session()

        # This is the collecter thread querying the metadata url
        self.ecs_metadata_thread = None

        # The fully qualified ARN for this process
        self._fq_arn = None

        # Response from the last call to
        # ${ECS_CONTAINER_METADATA_URI}/
        self.root_metadata = None

        # Response from the last call to
        # ${ECS_CONTAINER_METADATA_URI}/task
        self.task_metadata = None

        # Response from the last call to
        # ${ECS_CONTAINER_METADATA_URI}/stats
        self.stats_metadata = None

        # Response from the last call to
        # ${ECS_CONTAINER_METADATA_URI}/task/stats
        self.task_stats_metadata = None

        # Populate the collection helpers
        self.helpers.append(TaskHelper(self))
        self.helpers.append(DockerHelper(self))
        self.helpers.append(ProcessHelper(self))
        self.helpers.append(RuntimeHelper(self))
        self.helpers.append(ContainerHelper(self))

    def start(self):
        if self.ready_to_start is False:
            logger.warning("AWS Fargate Collector is missing requirements and cannot monitor this environment.")
            return

        super(AWSFargateCollector, self).start()

    def get_ecs_metadata(self):
        """
        Get the latest data from the ECS metadata container API and store on the class
        @return: Boolean
        """
        if env_is_test is True:
            # For test, we are using mock ECS metadata
            return

        try:
            delta = int(time()) - self.last_ecmu_full_fetch
            if delta > self.ecmu_full_fetch_interval:
                # Refetch the ECMU snapshot data
                self.last_ecmu_full_fetch = int(time())

                # Response from the last call to
                # ${ECS_CONTAINER_METADATA_URI}/
                json_body = self.http_client.get(self.ecmu_url_root, timeout=1).content
                self.root_metadata = json.loads(json_body)

                # Response from the last call to
                # ${ECS_CONTAINER_METADATA_URI}/task
                json_body = self.http_client.get(self.ecmu_url_task, timeout=1).content
                self.task_metadata = json.loads(json_body)

            # Response from the last call to
            # ${ECS_CONTAINER_METADATA_URI}/stats
            json_body = self.http_client.get(self.ecmu_url_stats, timeout=2).content
            self.stats_metadata = json.loads(json_body)

            # Response from the last call to
            # ${ECS_CONTAINER_METADATA_URI}/task/stats
            json_body = self.http_client.get(self.ecmu_url_task_stats, timeout=1).content
            self.task_stats_metadata = json.loads(json_body)
        except Exception:
            logger.debug("AWSFargateCollector.get_ecs_metadata", exc_info=True)

    def should_send_snapshot_data(self):
        delta = int(time()) - self.snapshot_data_last_sent
        if delta > self.snapshot_data_interval:
            return True
        return False

    def prepare_payload(self):
        payload = DictionaryOfStan()
        payload["spans"] = []
        payload["metrics"]["plugins"] = []

        try:
            if not self.span_queue.empty():
                payload["spans"] = self.queued_spans()

            with_snapshot = self.should_send_snapshot_data()

            # Fetch the latest metrics
            self.get_ecs_metadata()

            plugins = []
            for helper in self.helpers:
                plugins.extend(helper.collect_metrics(with_snapshot))

            payload["metrics"]["plugins"] = plugins

            if with_snapshot is True:
                self.snapshot_data_last_sent = int(time())
        except Exception:
            logger.debug("collect_snapshot error", exc_info=True)

        return payload

    def get_fq_arn(self):
        if self._fq_arn is not None:
            return self._fq_arn

        if self.root_metadata is not None:
            labels = self.root_metadata.get("Labels", None)
            if labels is not None:
                task_arn = labels.get("com.amazonaws.ecs.task-arn", "")

            container_name = self.root_metadata.get("Name", "")

            self._fq_arn = task_arn + "::" + container_name
            return self._fq_arn
        else:
            return "Missing ECMU metadata"



import uvicorn
from ...helpers import testenv
from instana.log import logger

testenv["starlette_port"] = 10817
testenv["starlette_server"] = ("http://127.0.0.1:" + str(testenv["starlette_port"]))

def launch_starlette():
    from .app import starlette_server
    from instana.singletons import agent

    # Hack together a manual custom headers list; We'll use this in tests
    agent.options.extra_http_headers = [u'X-Capture-This', u'X-Capture-That']

    uvicorn.run(starlette_server, host='127.0.0.1', port=testenv['starlette_port'], log_level="critical")



from __future__ import absolute_import

import opentracing
import opentracing.ext.tags as ext
import wrapt

from ..log import logger
from ..singletons import agent, tracer
from ..util import strip_secrets_from_query

try:
    import urllib3

    def collect(instance, args, kwargs):
        """ Build and return a fully qualified URL for this request """
        kvs = dict()
        try:
            kvs['host'] = instance.host
            kvs['port'] = instance.port

            if args is not None and len(args) == 2:
                kvs['method'] = args[0]
                kvs['path'] = args[1]
            else:
                kvs['method'] = kwargs.get('method')
                kvs['path'] = kwargs.get('path')
                if kvs['path'] is None:
                    kvs['path'] = kwargs.get('url')

            # Strip any secrets from potential query params
            if kvs.get('path') is not None and ('?' in kvs['path']):
                parts = kvs['path'].split('?')
                kvs['path'] = parts[0]
                if len(parts) == 2:
                    kvs['query'] = strip_secrets_from_query(parts[1], agent.options.secrets_matcher, agent.options.secrets_list)

            if type(instance) is urllib3.connectionpool.HTTPSConnectionPool:
                kvs['url'] = 'https://%s:%d%s' % (kvs['host'], kvs['port'], kvs['path'])
            else:
                kvs['url'] = 'http://%s:%d%s' % (kvs['host'], kvs['port'], kvs['path'])
        except Exception:
            logger.debug("urllib3 collect error", exc_info=True)
            return kvs
        else:
            return kvs

    def collect_response(scope, response):
        try:
            scope.span.set_tag(ext.HTTP_STATUS_CODE, response.status)

            if agent.options.extra_http_headers is not None:
                for custom_header in agent.options.extra_http_headers:
                    if custom_header in response.headers:
                        scope.span.set_tag("http.%s" % custom_header, response.headers[custom_header])

            if 500 <= response.status <= 599:
                scope.span.mark_as_errored()
        except Exception:
            logger.debug("collect_response", exc_info=True)

    @wrapt.patch_function_wrapper('urllib3', 'HTTPConnectionPool.urlopen')
    def urlopen_with_instana(wrapped, instance, args, kwargs):
        parent_span = tracer.active_span

        # If we're not tracing, just return; boto3 has it's own visibility
        if parent_span is None or parent_span.operation_name == 'boto3':
            return wrapped(*args, **kwargs)

        with tracer.start_active_span("urllib3", child_of=parent_span) as scope:
            try:
                kvs = collect(instance, args, kwargs)
                if 'url' in kvs:
                    scope.span.set_tag(ext.HTTP_URL, kvs['url'])
                if 'query' in kvs:
                    scope.span.set_tag("http.params", kvs['query'])
                if 'method' in kvs:
                    scope.span.set_tag(ext.HTTP_METHOD, kvs['method'])

                if 'headers' in kwargs:
                    tracer.inject(scope.span.context, opentracing.Format.HTTP_HEADERS, kwargs['headers'])

                response = wrapped(*args, **kwargs)

                collect_response(scope, response)

                return response
            except Exception as e:
                scope.span.mark_as_errored({'message': e})
                raise

    logger.debug("Instrumenting urllib3")
except ImportError:
    pass
